"""Glicko"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar
from riix.utils.constants import PI2, Q, Q2, Q2_3


class Glicko(OnlineRatingSystem):
    """
    Implements the original Glicko rating system, designed by Mark Glickman.

    This rating system is an improvement over the Elo rating system, introducing the concept of rating
    deviation and volatility to better account for the uncertainty in a player's true strength.
    """

    rating_dim = 2

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1500.0,
        initial_rating_dev: float = 350.0,
        c: float = 63.2,
        dtype=np.float64,
        update_method='iterative',
        # update_method='batched',
        do_weird_prob=False,
    ):
        """
        Initializes the Glicko rating system with the given parameters.

        Parameters:
            competitors (list): A list of competitors to be rated within the system.
            initial_rating (float, optional): The initial Glicko rating for new competitors. Defaults to 1500.0.
            initial_rating_dev (float, optional): The initial rating deviation for new competitors. Defaults to 350.0.
            c (float, optional): Constant used to adjust the rate of change of the rating deviation. Defaults to 63.2.
            dtype (data-type, optional): The desired data-type for the ratings and deviations arrays. Defaults to np.float64.
            update_method (str, optional): Method used for updating ratings ('iterative' or another specified method). Defaults to 'iterative'.
            do_weird_prob (bool, optional): If set to True, applies an alternative probability calculation. Defaults to False.
        """
        super().__init__(competitors)
        self.initial_rating_dev = initial_rating_dev
        self.c2 = c**2.0
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        self.rating_devs = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating_dev
        self.has_played = np.zeros(shape=self.num_competitors, dtype=np.bool_)
        self.prev_time_step = -1
        self.do_weird_prob = do_weird_prob

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    @staticmethod
    def g_vector(rating_dev):
        """
        Calculates the g function as part of the Glicko rating system.

        This function is used to scale the rating deviation, affecting the impact of a game outcome
        as a function of the opponent's rating volatility.

        Parameters:
            rating_dev (float): The rating deviation of an opponent.

        Returns:
            float: The calculated g function result, used to scale the expected score between players.
        """
        return 1.0 / np.sqrt(1.0 + (Q2_3 * np.square(rating_dev)) / PI2)

    @staticmethod
    def g_scalar(rating_dev):
        return 1.0 / math.sqrt(1.0 + (Q2_3 * (rating_dev) ** 2.0) / PI2)

    # TODO should Glicko probs incorporate the dev increase?
    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """generate predictions"""
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        rating_diffs = ratings_1 - ratings_2
        if self.do_weird_prob:
            # not sure why they do it this way but it seems to work better than the "real" way
            # https://github.com/McLeopold/PythonSkills/blob/95559262fbeaabc39cc5d698b93a6e43dc9b5e64/skills/glicko.py#L181
            probs = 1.0 / (1.0 + np.power(10, -rating_diffs / (2.0 * self.initial_rating_dev)))
        else:
            rating_devs_1 = self.rating_devs[matchups[:, 0]]
            rating_devs_2 = self.rating_devs[matchups[:, 1]]
            combined_dev = self.g_vector(np.sqrt(np.square(rating_devs_1) + np.square(rating_devs_2)))
            probs = sigmoid(Q * combined_dev * rating_diffs)
        return probs

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        means = self.ratings[matchups]
        devs = self.rating_devs[matchups]
        ratings = np.concatenate((means[..., None], devs[..., None]), axis=2).reshape(means.shape[0], -1)
        return ratings

    def increase_rating_dev(self, time_step, matchups):
        """called once per period to model the increase in variance over time"""
        active_in_period = np.unique(matchups)
        self.has_played[active_in_period] = True
        time_delta = time_step - self.prev_time_step
        self.rating_devs[self.has_played] = np.minimum(
            np.sqrt(np.square(self.rating_devs[self.has_played]) + (time_delta * self.c2)), self.initial_rating_dev
        )
        self.prev_time_step = time_step
        return active_in_period

    def batched_update(self, matchups, outcomes, time_step, use_cache=False, **kwargs):
        """apply one update based on all of the results of the rating period"""
        active_in_period = self.increase_rating_dev(time_step, matchups)
        masks = np.equal(matchups[:, :, None], active_in_period[None, :])  # N x 2 x active

        ratings = self.ratings[matchups]
        rating_diffs = ratings[:, 0] - ratings[:, 1]
        g_rating_devs = self.g_vector(self.rating_devs[matchups])
        probs_1 = sigmoid(Q * g_rating_devs[:, 1] * rating_diffs)
        probs_2 = sigmoid(-1.0 * (Q * g_rating_devs[:, 0] * rating_diffs))

        tmp = np.stack([probs_1 * (1.0 - probs_1), probs_2 * (1.0 - probs_2)]).T * np.square(g_rating_devs)[:, [1, 0]]
        d2 = 1.0 / ((tmp[:, :, None] * masks).sum(axis=(0, 1)) * Q2)

        outcomes = np.hstack([outcomes[:, None], 1.0 - outcomes[:, None]])
        probs = np.hstack([probs_1[:, None], probs_2[:, None]])

        r_num = Q * ((g_rating_devs[:, [1, 0]] * (outcomes - probs))[:, :, None] * masks).sum(axis=(0, 1))
        r_denom = (1.0 / np.square(self.rating_devs[active_in_period])) + (1.0 / d2)

        self.ratings[active_in_period] += r_num / r_denom
        self.rating_devs[active_in_period] = np.sqrt(1.0 / r_denom)

    def iterative_update(self, matchups, outcomes, time_step, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        self.increase_rating_dev(time_step, matchups)
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            rating_diff = self.ratings[comp_1] - self.ratings[comp_2]
            g_rating_devs = self.g_vector(self.rating_devs[matchups[idx]])
            g_rating_devs_2 = np.square(g_rating_devs)
            prob_1 = sigmoid_scalar(Q * g_rating_devs[1] * rating_diff)
            prob_2 = sigmoid_scalar(-Q * g_rating_devs[0] * rating_diff)
            d2_1 = 1.0 / (Q2 * prob_1 * (1.0 - prob_1) * g_rating_devs_2[1])
            d2_2 = 1.0 / (Q2 * prob_2 * (1.0 - prob_2) * g_rating_devs_2[0])
            r1_num = Q * g_rating_devs[1] * (outcomes[idx] - prob_1)
            r2_num = Q * g_rating_devs[0] * (1.0 - outcomes[idx] - prob_2)
            r1_denom = (1.0 / (self.rating_devs[comp_1] ** 2.0)) + (1.0 / d2_1)
            r2_denom = (1.0 / (self.rating_devs[comp_2] ** 2.0)) + (1.0 / d2_2)

            self.ratings[comp_1] += r1_num / r1_denom
            self.ratings[comp_2] += r2_num / r2_denom
            self.rating_devs[comp_1] = 1.0 / math.sqrt(r1_denom)
            self.rating_devs[comp_2] = 1.0 / math.sqrt(r2_denom)

    def print_leaderboard(self, num_places):
        sort_array = self.ratings - (3.0 * self.rating_devs)
        sorted_idxs = np.argsort(-sort_array)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating - (3*dev)"}\t')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{sort_array[comp_idx]:.6f}')



"""
variance incorporated Elo
Rating of players by Laplace approximation and dynamic modeling
https://arxiv.org/abs/2310.10386
"""
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid


class vElo(OnlineRatingSystem):
    """
    variance incorporated Elo by Weng, Lin, and their students!
    """

    rating_dim = 2

    def __init__(
        self,
        competitors: list,
        # initial_mu: float = 1500.0,
        # initial_sigma: float = 250.0,
        # sigma_reduction_factor: float = 1 / 5,  # A in the paper
        # sigma_lower_bound: float = 100.0,  # B in the paper
        # b: float = math.log(10.0) / 400.0,
        initial_mu: float = 0.0,
        initial_sigma: float = 1.2,
        sigma_reduction_factor: float = 1 / 5,  # A in the paper
        sigma_lower_bound: float = 0.4,  # B in the paper
        b: float = 1.0,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        super().__init__(competitors)
        self.mus = np.zeros(self.num_competitors, dtype=dtype) + initial_mu
        self.vs = np.zeros(self.num_competitors, dtype=dtype) + np.square(initial_sigma)
        self.sigma_reduction_factor = sigma_reduction_factor
        self.variance_lower_bound = sigma_lower_bound**2.0
        self.b = b
        self.b2 = b**2.0
        self.cache = {'probs': None}
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, time_step: int, matchups: np.ndarray, set_cache: bool = False):
        """generate predictions"""
        mus_1 = self.mus[matchups[:, 0]]
        mus_2 = self.mus[matchups[:, 1]]
        probs = sigmoid(self.b * (mus_1 - mus_2))
        if set_cache:
            self.cache['probs'] = probs
        return probs

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        means = self.mus[matchups]
        variances = self.vs[matchups]
        ratings = np.concatenate((means[..., None], variances[..., None]), axis=2).reshape(means.shape[0], -1)
        return ratings

    def batched_update(self, matchups, outcomes, use_cache):
        """apply one update based on all of the results of the rating period"""
        raise NotImplementedError

    def iterative_update(self, matchups, outcomes, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            outcome = outcomes[idx]
            mu_1, mu_2 = self.mus[matchups[idx]]
            v_1, v_2 = self.vs[matchups[idx]]
            prob = sigmoid(self.b * (mu_1 - mu_2))
            delta = outcome - prob

            C = 1.0 / (1.0 + self.b2 * prob * (1.0 - prob) * (v_1 + v_2))
            k_1 = v_1 * self.b * C
            k_2 = v_2 * self.b * C
            mu_1_p = mu_1 + (k_1 * delta)
            mu_2_p = mu_2 - (k_2 * delta)

            prob_p = sigmoid(self.b * (mu_1_p - mu_2_p))
            prob_prob = prob_p * (1.0 - prob_p)
            C_p = 1.0 / (1.0 + self.b2 * prob_prob * (v_1 + v_2))
            L_1 = prob_prob * v_1 * self.b2 * C_p
            L_2 = prob_prob * v_2 * self.b2 * C_p

            v_1_p = max(self.variance_lower_bound, v_1 * (1.0 - (self.sigma_reduction_factor * L_1)))
            v_2_p = max(self.variance_lower_bound, v_2 * (1.0 - (self.sigma_reduction_factor * L_2)))

            self.mus[comp_1] = mu_1_p
            self.mus[comp_2] = mu_2_p
            self.vs[comp_1] = v_1_p
            self.vs[comp_2] = v_2_p

    def print_leaderboard(self, num_places):
        sort_array = self.mus - (3.0 * np.sqrt(self.vs))
        sorted_idxs = np.argsort(-sort_array)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"mu - (3 * sd)"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{sort_array[comp_idx]:.6f}')



"""math utility functions for rating systems"""
import math
import statistics
import numpy as np
from scipy.special import expit
from scipy.stats import norm


def sigmoid(x):
    """a little faster than implementing it in numpy for d < 100000"""
    return expit(x)


def sigmoid_scalar(x):
    """no need to use numpy on scalars"""
    return 1.0 / (1.0 + math.exp(-x))


def base_10_sigmoid(x):
    """some methods prefer base 10 unfortunately"""
    return 1.0 / (1.0 + (10.0**-x))


INV_SQRT_2 = 1.0 / math.sqrt(2.0)


def norm_cdf(x):
    """cdf of standard normal"""
    return 0.5 * (1.0 + math.erf(x * INV_SQRT_2))


STANDARD_NORMAL = statistics.NormalDist()


def norm_pdf(x):
    """pdf of standard normal"""
    return STANDARD_NORMAL.pdf(x)


def v_and_w_win_vector(t, eps):
    """calculate v and w for a win in a vectorized fashion"""
    diff = t - eps
    v = np.empty_like(diff)
    denom = norm.cdf(diff)
    bad_mask = denom < 1e-50
    v[bad_mask] = -1.0 * diff[bad_mask]
    v[~bad_mask] = norm.pdf(diff[~bad_mask]) / denom[~bad_mask]
    w = v * (v + diff)
    return v, w


def v_and_w_draw_vector(t, eps):
    """calculate v and w for a draw in a vectorized fashion"""
    abs_t = np.abs(t)  # the papers do NOT do this but ALL open source implementations DO...
    diff_a = eps - abs_t
    diff_b = -eps - abs_t

    # TODO maybe this would be faster if I concatenated and put through pdf/cdf together?
    pdf_a = norm.pdf(diff_a)
    pdf_b = norm.pdf(diff_b)
    cdf_a = norm.cdf(diff_a)
    cdf_b = norm.cdf(diff_b)

    v_num = pdf_b - pdf_a
    shared_denom = cdf_a - cdf_b
    bad_mask = shared_denom < 1e-5
    good_mask = ~bad_mask

    v = np.empty_like(t)
    if eps.shape != t.shape:
        eps = np.repeat(eps, repeats=2, axis=1)
    v[bad_mask] = -t[bad_mask] + (np.sign(t[bad_mask]) * eps[bad_mask])
    v[good_mask] = v_num[good_mask] / shared_denom[good_mask]

    w = np.empty_like(t)
    w_bad_mask = np.isnan(shared_denom) | np.isinf(shared_denom) | (shared_denom < 1e-50)
    w_good_mask = ~w_bad_mask
    w[w_bad_mask] = 1.0

    w_num = (diff_a[w_good_mask] * pdf_a[w_good_mask]) - (diff_b[w_good_mask] * pdf_b[w_good_mask])
    w[w_good_mask] = (w_num / shared_denom[w_good_mask]) + np.square(v[w_good_mask])
    return v, w


def v_and_w_win_scalar(t, eps):
    """calculate v and w for a win in a scalar fashion"""
    diff = t - eps
    cdf = norm_cdf(diff)
    if cdf > 2.222758749e-162:
        v = norm_pdf(diff) / cdf
    else:
        v = -diff
    w = v * (v + diff)
    return v, w


def v_and_w_draw_scalar(t, eps):
    """calculate v and w for a draw in a scalar fashion"""
    abs_t = math.fabs(t)  # the papers do NOT do this but ALL open source implementations DO...
    diff_a = eps - abs_t
    diff_b = -eps - abs_t

    cdf_a = norm_cdf(diff_a)
    cdf_b = norm_cdf(diff_b)

    pdf_a = norm_pdf(diff_a)
    pdf_b = norm_pdf(diff_b)
    v_num = pdf_a - pdf_b
    shared_denom = cdf_a - cdf_b
    sign = math.copysign(1.0, t)
    if shared_denom < 1e-5:
        v = -t + (sign * eps)
    else:
        v = sign * v_num / shared_denom
    if shared_denom < 1e-50:
        w = 1.0
    else:
        w_num = (diff_a * pdf_a) - (diff_b * pdf_b)
        w = math.copysign(1.0, t) * ((w_num / shared_denom) + (v**2.0))
    return v, w



"""The Elo-Davidson rating system"""
import numpy as np
import math
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid


class EloDavidson(OnlineRatingSystem):
    """
    Implements the Elo-Davidson rating system from https://www.researchgate.net/publication/341384358_Understanding_Draws_in_Elo_Rating_Algorithm
    This method applies the method of handling draws proposed by Davidson to the "online" Elo rating system
    Davidson's paper: https://www.jstor.org/stable/2283595
    """

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 0.0,  # this does nothing
        k: float = 32.0,
        kappa: float = 1.0,  # kappa = 2, sigma = 200 is equivalent to Elo with scale = 400.0
        base: float = 10.0,
        sigma: float = 200.0,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        """
        Initializes the Elo rating system with the given parameters.

        Parameters:
            competitors (list): A list of competitors to be rated within the system.
            initial_rating (float, optional): The initial Elo rating for new competitors. Defaults to 1500.0.
        """
        super().__init__(competitors)
        self.k = k
        self.kappa = kappa
        self.kappa_over_2 = kappa / 2.0
        self.alpha = math.log(base) / (2.0 * sigma)
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        Parameters:
            matchups (np.ndarray): A NumPy array of matchups, where each row represents a matchup
                                    and contains two integers indicating the indices of the competitors
                                    in the 'ratings' array.
            time_step (int, optional): A time step at which the predictions are made. This parameter
                                    is not used in the current implementation but can be utilized
                                    for time-dependent predictions. Defaults to None.
            set_cache (bool, optional): If True, caches the computed probabilities in the 'cache'
                                        attribute under the key 'probs'. Defaults to False.

        Returns:
            np.ndarray: A NumPy array containing the predicted probabilities for the first competitor
                        in each matchup winning against the second.
        """
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        probs = sigmoid(self.alpha * (ratings_1 - ratings_2))
        return probs

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.ratings[matchups]

    def batched_update(self, matchups, outcomes, use_cache):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
            use_cache: Flag to use cached probabilities or calculate anew.
        """
        raise NotImplementedError

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
            **kwargs: Additional parameters (not used).
        """
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            r_1 = self.ratings[comp_1]
            r_2 = self.ratings[comp_2]

            diff = math.exp(self.alpha * (r_1 - r_2))
            denom = diff + (1.0 / diff) + self.kappa

            p_1 = (diff + self.kappa_over_2) / denom
            p_2 = ((1.0 / diff) + self.kappa_over_2) / denom

            update_1 = self.k * (outcomes[idx] - p_1)
            update_2 = self.k * (1.0 - outcomes[idx] - p_2)

            self.ratings[comp_1] += update_1
            self.ratings[comp_2] += update_2

    def print_leaderboard(self, num_places):
        sorted_idxs = np.argsort(-self.ratings)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{self.ratings[comp_idx]:.6f}')



"""The Elo rating system"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar


class Elo(OnlineRatingSystem):
    """
    Implements the original Elo rating system.
    """

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1500.0,
        k: float = 32.0,
        alpha: float = math.log(10.0) / 400.0,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        """
        Initializes the Elo rating system with the given parameters.

        Parameters:
            competitors (list): A list of competitors to be rated within the system.
            initial_rating (float, optional): The initial Elo rating for new competitors. Defaults to 1500.0.
            k (float, optional): The K-factor, which controls the rate at which ratings change. Defaults to 32.0.
            alpha (float, optional): Scaling factor used in the calculation of expected scores. Defaults to log(10) / 400.
            update_method (str, optional): Method used to update ratings ('iterative' or other methods if implemented). Defaults to 'iterative'.
            dtype: The data type for internal numpy computations. Defaults to np.float64.

        Initializes an Elo rating system with customizable settings for initial ratings, K-factor, and update method.
        """
        super().__init__(competitors)
        self.k = k
        self.alpha = alpha
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        self.cache = {'probs': None}
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.

        This method calculates the probability of the first competitor in each matchup winning
        based on their Elo ratings. The probabilities are computed using the sigmoid function
        applied to the rating differences, scaled by the alpha parameter.

        Parameters:
            matchups (np.ndarray): A NumPy array of matchups, where each row represents a matchup
                                    and contains two integers indicating the indices of the competitors
                                    in the 'ratings' array.
            time_step (int, optional): A time step at which the predictions are made. This parameter
                                    is not used in the current implementation but can be utilized
                                    for time-dependent predictions. Defaults to None.
            set_cache (bool, optional): If True, caches the computed probabilities in the 'cache'
                                        attribute under the key 'probs'. Defaults to False.

        Returns:
            np.ndarray: A NumPy array containing the predicted probabilities for the first competitor
                        in each matchup winning against the second.
        """
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        probs = sigmoid(self.alpha * (ratings_1 - ratings_2))
        if set_cache:
            self.cache['probs'] = probs
        return probs

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.ratings[matchups]

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
            use_cache: Flag to use cached probabilities or calculate anew.
        """
        active_in_period = np.unique(matchups)
        masks = np.equal(matchups[:, :, None], active_in_period[None, :])  # N x 2 x active
        if use_cache:
            probs = self.cache['probs']
        else:
            probs = self.predict(time_step=None, matchups=matchups, set_cache=False)
        per_match_diff = (outcomes - probs)[:, None]
        per_match_diff = np.hstack([per_match_diff, -per_match_diff])
        per_competitor_diff = (per_match_diff[:, :, None] * masks).sum(axis=(0, 1))
        self.ratings[active_in_period] += self.k * per_competitor_diff

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
            **kwargs: Additional parameters (not used).
        """
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            diff = self.ratings[comp_1] - self.ratings[comp_2]
            prob = sigmoid_scalar(self.alpha * diff)
            update = self.k * (outcomes[idx] - prob)
            self.ratings[comp_1] += update
            self.ratings[comp_2] -= update

    def print_leaderboard(self, num_places):
        sorted_idxs = np.argsort(-self.ratings)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{self.ratings[comp_idx]:.6f}')



"""
example from: https://github.com/sublee/trueskill/blob/master/trueskilltest.py
"""
import math
import pytest
import numpy as np
from riix.models.trueskill import TrueSkill
from riix.utils.data_utils import MatchupDataset


def general_trueskill(update_method):
    time_steps = np.array([0, 0])
    matchups = np.array([[0, 1], [2, 3]])
    outcomes = np.array([1.0, 0.5])
    dataset = MatchupDataset.init_from_arrays(
        time_steps=time_steps, matchups=matchups, outcomes=outcomes, competitors=[0, 1, 2, 3]
    )
    model = TrueSkill(
        competitors=dataset.competitors,
        initial_mu=25.0,
        initial_sigma=25.0 / 3.0,
        beta=25.0 / 6.0,
        draw_probability=0.1,
        update_method=update_method,
    )
    model.fit_dataset(dataset)
    return model


def test_trueskill_iterative():
    model = general_trueskill('iterative')
    assert model.mus[0] == pytest.approx(29.396, rel=1e-4)
    assert model.mus[1] == pytest.approx(20.604, rel=1e-4)
    assert math.sqrt(model.sigma2s[0]) == pytest.approx(7.171, rel=1e-4)
    assert math.sqrt(model.sigma2s[1]) == pytest.approx(7.171, rel=1e-4)

    assert model.mus[2] == pytest.approx(25.0, rel=1e-4)
    assert model.mus[3] == pytest.approx(25.0, rel=1e-4)
    assert math.sqrt(model.sigma2s[2]) == pytest.approx(6.458, rel=5e-4)
    assert math.sqrt(model.sigma2s[3]) == pytest.approx(6.458, rel=5e-4)


def test_trueskill_batched():
    model = general_trueskill('batched')
    assert model.mus[0] == pytest.approx(29.396, rel=1e-4)
    assert model.mus[1] == pytest.approx(20.604, rel=1e-4)
    assert math.sqrt(model.sigma2s[0]) == pytest.approx(7.171, rel=1e-4)
    assert math.sqrt(model.sigma2s[1]) == pytest.approx(7.171, rel=1e-4)

    assert model.mus[2] == pytest.approx(25.0, rel=1e-4)
    assert model.mus[3] == pytest.approx(25.0, rel=1e-4)
    assert math.sqrt(model.sigma2s[2]) == pytest.approx(6.458, rel=5e-4)
    assert math.sqrt(model.sigma2s[3]) == pytest.approx(6.458, rel=5e-4)



"""a template to copy and paste when implementing a new rating system"""
import numpy as np
from riix.core.base import OnlineRatingSystem


class Template(OnlineRatingSystem):
    """put a docstring"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        argument: str = 'lorem ipsum',
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        """Initialize the rating system"""
        super().__init__(competitors)
        self.argument = argument
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        """
        return np.zeros(matchups.shape[0]) + 0.5

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        raise NotImplementedError

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
        """
        raise NotImplementedError

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
        """
        raise NotImplementedError

    def print_leaderboard(self, num_places):
        raise NotImplementedError



"""
Online version of the Disc Decomposition rating system
https://proceedings.mlr.press/v206/bertrand23a.html
equations 8 and 9
"""
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid


class OnlineDiscDecomp(OnlineRatingSystem):
    """Online Disc Decomposition"""

    rating_dim = 2

    def __init__(
        self,
        competitors: list,
        initial_u: float = 1.0,
        initial_v: float = 1.0,
        eta: float = 0.02,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        super().__init__(competitors)
        self.initial_u = initial_u
        self.initial_v = initial_v
        self.eta = eta
        self.us = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_u
        self.vs = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_v
        self.cache = {'probs': None}
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        us = self.us[matchups]
        vs = self.vs[matchups]
        ratings = np.concatenate((us[..., None], vs[..., None]), axis=2).reshape(us.shape[0], -1)
        return ratings

    def predict(self, matchups: np.ndarray, set_cache: bool = False, **kwargs):
        us_1 = self.us[matchups[:, 0]]
        us_2 = self.us[matchups[:, 1]]
        vs_1 = self.vs[matchups[:, 0]]
        vs_2 = self.vs[matchups[:, 1]]
        probs = sigmoid(vs_1 * vs_2 * (us_1 - us_2))
        if set_cache:
            self.cache['probs'] = probs
        return probs

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        active_in_period = np.unique(matchups)
        masks = np.equal(matchups[:, :, None], active_in_period[None, :])  # N x 2 x active
        if use_cache:
            probs = self.cache['probs']
        else:
            probs = self.predict(time_step=None, matchups=matchups, set_cache=False)
        per_match_diff = (outcomes - probs)[:, None]
        per_match_diff = np.hstack([per_match_diff, -per_match_diff])
        per_competitor_diff = (per_match_diff[:, :, None] * masks).sum(axis=(0, 1))
        self.ratings[active_in_period] += self.k * per_competitor_diff

    def iterative_update(self, matchups, outcomes, **kwargs):
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            u_1, u_2 = self.us[matchups[idx]]
            v_1, v_2 = self.vs[matchups[idx]]
            u_diff = u_1 - u_2
            v_prod = v_1 * v_2
            prob = sigmoid(v_prod * u_diff)
            scaled_delta = self.eta * (outcomes[idx] - prob)

            u_update = scaled_delta * v_prod
            v_update = scaled_delta * u_diff

            self.us[comp_1] += u_update
            self.us[comp_2] -= u_update
            self.us[comp_1] += v_update * v_2
            self.us[comp_2] -= v_update * v_1

    def print_leaderboard(self, num_places):
        sorted_idxs = np.argsort(-self.us)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{self.us[comp_idx]:.6f}')



"""
Constant Variance Glicko
Martin Ingram 2021
https://www.researchgate.net/publication/348511584_How_to_extend_Elo_a_Bayesian_perspective
"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar


class ConstantVarianceGlicko(OnlineRatingSystem):
    """Glicko but all competitors have equal and unchanging variance"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1500.0,
        rd: float = 81.7,  # for tennis Ingram found this should be 81.7
        b: float = math.log(10.0) / 400.0,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        """Initialize the rating system"""
        super().__init__(competitors)
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        self.rd = rd
        self.b = b
        self.b2 = b**2.0

        # a whole bunch of stuff is constant :)
        self.rd2_inv = 1.0 / (rd**2.0)
        self.g_rd = 1.0 / math.sqrt(1.0 + (3.0 * (self.b2) * (rd**2.0)) / (math.pi**2.0))
        self.b_g_rd = b * self.g_rd
        self.b2_g_rd2 = self.b_g_rd**2.0

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        """
        ratings = self.ratings[matchups]
        return sigmoid(self.b_g_rd * (ratings[:, 0] - ratings[:, 1]))

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        raise NotImplementedError

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
        """
        raise NotImplementedError

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
        """
        # somebody on twitter told me .self was slow so we are getting these once first
        rd2_inv = self.rd2_inv
        b_g_rd = self.b_g_rd
        b2_g_rd2 = self.b2_g_rd2
        for matchup, outcome in zip(matchups, outcomes):
            r_1, r_2 = self.ratings[matchup]
            prob = sigmoid_scalar(b_g_rd * (r_1 - r_2))
            d_squared_inv = b2_g_rd2 * prob * (1.0 - prob)
            k = b_g_rd / (rd2_inv + d_squared_inv)
            delta = k * (outcome - prob)
            self.ratings[matchup[0]] += delta
            self.ratings[matchup[1]] -= delta

    def print_leaderboard(self, num_places):
        raise NotImplementedError



"""Weng/Lin Bayesian Online Rating system"""
import math
import numpy as np
from scipy.stats import norm
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar
from riix.utils.math_utils import v_and_w_win_scalar, v_and_w_draw_scalar


class WengLin(OnlineRatingSystem):
    """The Bayesian Online Rating System introduced by Weng and Lin"""

    rating_dim = 2

    def __init__(
        self,
        competitors: list,
        initial_mu: float = 25.0,
        initial_sigma: float = 8.333,
        beta: float = 4.166,
        kappa: float = 0.0001,
        tau: float = 0.0833,
        draw_probability=0.0,
        update_method: str = 'iterative',
        model='tm',
        dtype=np.float64,
    ):
        super().__init__(competitors)
        self.beta = beta
        self.two_beta_squared = 2.0 * (beta**2.0)
        self.tau_squared = tau**2.0
        self.kappa = kappa
        self.epsilon = norm.ppf((draw_probability + 1.0) / 2.0) * math.sqrt(2.0) * beta

        self.mus = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_mu
        self.sigma2s = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_sigma**2.0
        self.has_played = np.zeros(shape=self.num_competitors, dtype=np.bool_)
        self.prev_time_step = 0

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

        if model == 'bt':
            self.model_func = self.bradley_terry_scalar_updates
            self.prob_func = sigmoid
        elif model == 'tm':
            self.model_func = self.thurstone_mosteller_scalar_updates
            self.prob_func = norm.cdf

        self.cache = {
            'combined_sigma2s': None,
            'combined_devs': None,
            'probs': None,
        }

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        mus = self.mus[matchups]
        sigma2s = self.sigma2s[matchups]
        ratings = np.concatenate((mus[..., None], sigma2s[..., None]), axis=2).reshape(mus.shape[0], -1)
        return ratings

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """generate predictions"""
        # TODO should add time based var increase to prediction logic
        mus = self.mus[matchups]
        sigma2s = self.sigma2s[matchups]
        combined_sigma2s = self.two_beta_squared + sigma2s.sum(axis=1)
        combined_devs = np.sqrt(combined_sigma2s)
        norm_diffs = (mus[:, 0] - mus[:, 1]) / combined_devs
        probs = self.prob_func(norm_diffs)
        if set_cache:
            self.cache['combined_sigma2s'] = combined_sigma2s
            self.cache['combined_devs'] = combined_devs
            self.cache['probs'] = probs
        return probs

    def increase_rating_dev(self, time_step, matchups):
        """called once per period to model the increase in variance over time"""
        active_in_period = np.unique(matchups)
        self.has_played[active_in_period] = True
        time_delta = time_step - self.prev_time_step
        self.sigma2s[self.has_played] += time_delta * self.tau_squared  # increase var for active players
        self.prev_time_step = time_step
        return active_in_period

    def batched_update(self, matchups, outcomes, use_cache=False, **kwargs):
        """apply one update based on all of the results of the rating period"""
        # TODO fix the update to actually be based on time...
        active_in_period = self.increase_rating_dev(matchups)
        masks = np.equal(matchups[:, :, None], active_in_period[None, :])  # N x 2 x active

        mus = self.mus[matchups]
        sigma2s = self.sigma2s[matchups]

        if use_cache:
            combined_sigma2s = self.cache['combined_sigma2s']
            combined_devs = self.cache['combined_devs']
            probs = self.cache['probs']
        else:
            combined_sigma2s = self.two_beta_squared + sigma2s.sum(axis=1)
            combined_devs = np.sqrt(combined_sigma2s)
            norm_diffs = (mus[:, 0] - mus[:, 1]) / combined_devs
            probs = sigmoid(norm_diffs)

        mu_updates = (sigma2s / combined_devs[:, None]) * (outcomes - probs)[:, None]
        mu_updates[:, 1] *= -1.0

        gammas = np.sqrt(sigma2s) / combined_devs[:, None]
        etas = gammas * (sigma2s / combined_sigma2s[:, None]) * (probs * (1.0 - probs))[:, None]

        sigma2_multipliers = np.maximum(1.0 - etas, self.kappa)

        mu_updates_pooled = (mu_updates[:, :, None] * masks).sum((0, 1))
        # aggregating by prod does NOT work well empirically, variance decreases way too fast
        # empirically max works really well
        sigma2_multipliers_pooled = np.max(sigma2_multipliers[:, :, None] * masks, axis=(0, 1))

        self.mus[active_in_period] += mu_updates_pooled
        self.sigma2s[active_in_period] *= sigma2_multipliers_pooled

    def bradley_terry_scalar_updates(self, norm_diff, sigma2s, combined_sigma2, combined_dev, outcome):
        prob = sigmoid_scalar(norm_diff)
        deltas = (sigma2s / combined_dev) * (outcome - prob)
        gammas = np.sqrt(sigma2s) / combined_dev
        etas = gammas * (sigma2s / combined_sigma2) * (prob * (1.0 - prob))
        return deltas, etas

    def thurstone_mosteller_scalar_updates(self, norm_diff, sigma2s, combined_sigma2, combined_dev, outcome):
        sign_multiplier = outcome if outcome else -1.0
        if outcome != 0.5:
            v, w = v_and_w_win_scalar(norm_diff * sign_multiplier, self.epsilon / combined_dev)
        else:
            v, w = v_and_w_draw_scalar(norm_diff, self.epsilon / combined_dev)

        deltas = sign_multiplier * (sigma2s / combined_dev) * v
        gammas = np.sqrt(sigma2s) / combined_dev
        etas = gammas * (sigma2s / combined_sigma2) * w
        return deltas, etas

    def iterative_update(self, matchups, outcomes, time_step, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        self.increase_rating_dev(time_step, matchups)
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            # self.sigma2s[matchups[idx]] += self.tau_squared
            sigma2s = self.sigma2s[matchups[idx]]
            combined_sigma2 = self.two_beta_squared + sigma2s.sum()
            combined_dev = np.sqrt(combined_sigma2)
            norm_diff = (self.mus[comp_1] - self.mus[comp_2]) / combined_dev

            deltas, etas = self.model_func(norm_diff, sigma2s, combined_sigma2, combined_dev, outcomes[idx])
            sigma2_multipliers = np.maximum(1.0 - etas, self.kappa)

            self.mus[comp_1] += deltas[0]
            self.mus[comp_2] -= deltas[1]
            self.sigma2s[matchups[idx]] *= sigma2_multipliers

    def print_leaderboard(self, num_places):
        sort_array = self.mus - (3.0 * np.sqrt(self.sigma2s))
        sorted_idxs = np.argsort(-sort_array)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"mu - (3 * sd)"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{sort_array[comp_idx]:.6f}')



"""
Glicko 2
paper: http://www.glicko.net/research/dpcmsv.pdf
example: http://www.glicko.net/glicko/glicko2.pdf

"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar
from riix.utils.constants import THREE_OVER_PI_SQUARED


class Glicko2(OnlineRatingSystem):
    """
    Implements the Glicko 2 rating system, designed by Mark Glickman.
    """

    rating_dim = 2

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1500.0,
        initial_rd: float = 350.0,
        initial_sigma: float = 0.06,
        tau: float = 0.5,
        epsilon: float = 1e-6,
        dtype=np.float64,
        # update_method='batched',
        update_method='iterative',
    ):
        """Initializes the Glicko rating system with the given parameters."""
        super().__init__(competitors)
        self.mus = np.zeros(shape=self.num_competitors, dtype=dtype) + ((initial_rating - 1500.0) / 173.7178)
        self.initial_phi = initial_rd / 173.7178

        self.phis = np.zeros(shape=self.num_competitors, dtype=dtype) + self.initial_phi
        self.sigmas = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_sigma
        self.has_played = np.zeros(shape=self.num_competitors, dtype=np.bool_)
        self.prev_time_step = -1
        self.tau = tau
        self.tau2 = tau**2.0
        self.epsilon = epsilon

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    @staticmethod
    def g_scalar(phi):
        """this is DIFFERENT from g in regular Glicko"""
        return 1.0 / math.sqrt(1.0 + (THREE_OVER_PI_SQUARED * (phi**2.0)))

    @staticmethod
    def g_vector(phi):
        """vector version"""
        return 1.0 / np.sqrt(1.0 + (THREE_OVER_PI_SQUARED * np.square(phi)))

    # TODO should Glicko probs incorporate the dev increase?
    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """generate predictions"""
        mu_1 = self.mus[matchups[:, 0]]
        mu_2 = self.mus[matchups[:, 1]]
        mu_diff = mu_1 - mu_2
        phi_1 = self.phis[matchups[:, 0]]
        phi_2 = self.phis[matchups[:, 1]]
        # the papers and theory seem to indicate this...
        combined_phi = self.g_vector(np.sqrt(np.square(phi_1) + np.square(phi_2)))
        # but this seems to work better...
        # combined_phi = self.g_vector(phi_1 + phi_2)
        probs = sigmoid(combined_phi * mu_diff)
        return probs

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        means = self.mus[matchups]
        devs = self.phis[matchups]
        ratings = np.concatenate((means[..., None], devs[..., None]), axis=2).reshape(means.shape[0], -1)
        return ratings

    def increase_rating_dev(self, time_step, matchups):
        """called once per period to model the increase in variance over time"""
        active_in_period = np.unique(matchups)
        self.has_played[active_in_period] = True
        time_delta = time_step - self.prev_time_step
        self.phis[self.has_played] = np.minimum(
            np.sqrt(np.square(self.phis[self.has_played]) + (time_delta * np.square(self.sigmas[self.has_played]))),
            self.initial_phi,
        )
        self.prev_time_step = time_step
        return active_in_period

    def batched_update(self, matchups, outcomes, time_step, **kwargs):
        """apply one update based on all of the results of the rating period"""
        active_in_period = np.unique(matchups)  # (n_active,)
        self.has_played[active_in_period] = True
        # update phi for players who have played but are not active in this rating period
        inactive_mask = np.ones(self.num_competitors, dtype=np.bool_)
        inactive_mask[active_in_period] = False
        update_phi_mask = self.has_played & inactive_mask
        time_delta = time_step - self.prev_time_step
        self.phis[update_phi_mask] = np.minimum(
            np.sqrt(np.square(self.phis[update_phi_mask]) + (time_delta * np.square(self.sigmas[update_phi_mask]))),
            self.initial_phi,
        )

        active_mask = np.equal(matchups[:, :, None], active_in_period[None, :])  # (M,2,n_active)
        mus = self.mus[matchups]
        phis = self.phis[matchups]
        gs = self.g_vector(phis)[:, [1, 0]]  # swap gs since competitors interact with opponents' variance
        mu_diffs = mus - mus[:, [1, 0]]
        probs = sigmoid(gs * mu_diffs)
        stacked_outcomes = np.column_stack((outcomes, 1.0 - outcomes))

        vs = 1.0 / ((np.square(gs) * probs * (1.0 - probs))[:, :, None] * active_mask).sum(axis=(0, 1))  # (n_active,)
        # this is kinda like a gradient
        grads = ((gs * (stacked_outcomes - probs))[:, :, None] * active_mask).sum(axis=(0, 1))  # (n_active,)
        deltas = vs * grads

        sigma_primes = np.empty(active_in_period.shape)
        for idx, comp in enumerate(active_in_period):
            sigma_primes[idx] = self.get_sigma_prime(
                phi=self.phis[comp], delta=deltas[idx], v=vs[idx], sigma=self.sigmas[comp]
            )

        phi_stars_squared = np.square(self.phis[active_in_period]) + (time_delta * np.square(sigma_primes))
        phi_primes = 1.0 / np.sqrt((1.0 / phi_stars_squared) + (1.0 / vs))

        mu_updates = np.square(phi_primes) * grads

        self.mus[active_in_period] += mu_updates
        self.phis[active_in_period] = phi_primes
        self.sigmas[active_in_period] = sigma_primes
        self.prev_time_step = time_step

    def f(self, x, delta2, phi2, v, a):
        ex = math.exp(x)
        phi2_v_ex = phi2 + v + ex
        num_1 = ex * (delta2 - phi2_v_ex)
        denom_1 = 2 * ((phi2_v_ex) ** 2.0)
        term_2 = (x - a) / self.tau2
        return (num_1 / denom_1) - term_2

    def get_sigma_prime(self, phi, delta, v, sigma):
        delta2 = delta**2.0
        phi2 = phi**2.0
        A = a = math.log(sigma**2.0)
        if delta2 > (phi2 + v):
            B = math.log(delta2 - phi2 - v)
        else:
            B = a - self.tau
            while self.f(B, delta2=delta2, phi2=phi2, v=v, a=a) < 0:
                B -= self.tau

        f_A = self.f(A, delta2, phi2, v, a)
        f_B = self.f(B, delta2, phi2, v, a)
        while math.fabs(A - B) > self.epsilon:
            C = A + ((A - B) * f_A) / (f_B - f_A)
            f_C = self.f(C, delta2, phi2, v, a)
            if (f_C * f_B) <= 0:
                A = B
                f_A = f_B
            else:
                f_A = f_A / 2.0
            B = C
            f_B = f_C
        sigma_prime = math.exp(A / 2.0)
        return sigma_prime

    def iterative_update(self, matchups, outcomes, time_step, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        # increase the phi's once at the beginning of the rating period with the prior sigma
        self.increase_rating_dev(time_step, matchups)
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            mu_1 = self.mus[comp_1]
            mu_2 = self.mus[comp_2]
            mu_diff = mu_1 - mu_2
            phi_1 = self.phis[comp_1]
            phi_2 = self.phis[comp_2]
            g_1 = self.g_scalar(phi_1)
            g_2 = self.g_scalar(phi_2)
            p_1 = sigmoid_scalar(g_2 * mu_diff)
            p_2 = sigmoid_scalar(-g_1 * mu_diff)
            v_1 = 1.0 / ((g_2**2.0) * p_1 * (1.0 - p_1))
            v_2 = 1.0 / ((g_1**2.0) * p_2 * (1.0 - p_2))

            delta_1 = v_1 * g_2 * (outcomes[idx] - p_1)
            delta_2 = v_2 * g_1 * (1.0 - outcomes[idx] - p_2)

            sigma_star_1 = self.get_sigma_prime(phi_1, delta_1, v_1, self.sigmas[comp_1])
            sigma_star_2 = self.get_sigma_prime(phi_2, delta_2, v_2, self.sigmas[comp_2])

            # update the sigmas for each match
            self.sigmas[comp_1] = sigma_star_1
            self.sigmas[comp_2] = sigma_star_2

            # update the phis in each match
            # we do not need to update the phis for the passage of time during each match, that's done before the loop
            self.phis[comp_1] = 1.0 / math.sqrt((1.0 / (phi_1**2.0)) + (1.0 / v_1))
            self.phis[comp_2] = 1.0 / math.sqrt((1.0 / (phi_2**2.0)) + (1.0 / v_2))

            self.mus[comp_1] += (self.phis[comp_1] ** 2.0) * g_2 * (outcomes[idx] - p_1)
            self.mus[comp_2] += (self.phis[comp_2] ** 2.0) * g_1 * (1.0 - outcomes[idx] - p_2)

    def print_leaderboard(self, num_places):
        sort_array = self.mus - (3.0 * np.square(self.phis))
        sorted_idxs = np.argsort(-sort_array)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating - (3*dev)"}\t')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{sort_array[comp_idx]:.6f}')



"""Multidimensional Elo"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.models.elo import Elo
from riix.utils.math_utils import sigmoid


def generate_orthogonal_matrix(d, k):
    # Constraint check
    if d > k + 1:
        raise ValueError('d must be less than or equal to k + 1')

    # Initialize the matrix with zeros
    matrix = np.zeros((d, k))

    # Generate a base pattern
    base_pattern = np.array([1.0, -1.0] * (k // 2) + [1.0] * (k % 2))

    # Fill the first d-1 rows
    for i in range(d - 1):
        shift = i % k  # Calculate the shift for the pattern
        matrix[i] = np.roll(base_pattern, shift)

    # Fill the d-th row
    matrix[-1] = -np.sum(matrix[:-1], axis=0)
    return matrix


class Melo(Elo):
    """Multidimensional Elo rating system, (good for rock paper scissors problems)"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1500.0,
        dimension: int = 1,  # this is the k in melo_2k, not sure why they had to use that letter when it's already used in Elo smh
        eta_r: float = 32.0,  # this is the normal elo k factor
        eta_c: float = 0.125,  # 1 is bad: https://dclaz.github.io/mELO/articles/03_noise.html#simulations-1
        alpha: float = math.log(10.0) / 400.0,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        OnlineRatingSystem.__init__(self, competitors)
        self.eta_r = eta_r
        self.eta_c = eta_c
        self.alpha = alpha
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        two_k = 2 * dimension
        self.c = generate_orthogonal_matrix(d=two_k, k=self.num_competitors).T # / 10.0

        # rng = np.random.default_rng(42)
        # self.c = rng.uniform(low=-1.0, high=1.0, size=(num_competitors, two_k))

        self.omega = np.zeros((two_k, two_k), dtype=np.int32)
        # Set every other off-diagonal element to 1 or -1
        self.omega[np.arange(0, two_k - 1, 2), np.arange(1, two_k, 2)] = 1  # Upper off diagonal
        self.omega[np.arange(1, two_k, 2), np.arange(0, two_k - 1, 2)] = -1  # Lower off diagonal

        if update_method == 'batched':
            self.update = self.batched_update
        if update_method == 'iterative':
            self.update = self.iterative_update

        self.cache = {'probs': None, 'c_1_times_omega': None}

    def get_pre_match_ratings(self, matchups, **kwargs):
        return self.ratings[matchups]

    def predict(self, matchups: np.ndarray, set_cache: bool = False, **kwargs):
        """generate predictions"""
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        c_1 = self.c[matchups[:, 0], None, :]  # [bs, 1, 2k]
        c_2 = self.c[matchups[:, 1], :, None]  # [bs, 2k, 1]
        elo_diff = ratings_1 - ratings_2
        c_1_times_omega = np.matmul(c_1, self.omega)
        melo_diff = np.matmul(c_1_times_omega, c_2)[:, 0, 0]
        probs = sigmoid(self.alpha * (elo_diff + melo_diff))
        if set_cache:
            self.cache['probs'] = probs
            self.cache['c_1_times_omega'] = c_1_times_omega.squeeze(axis=1)

        return probs

    def batched_update(self, matchups, outcomes, time_step=None, use_cache=False):
        """apply one update based on all of the results of the rating period"""
        active_in_period = np.unique(matchups)
        masks = np.equal(matchups[:, :, None], active_in_period[None, :])  # N x 2 x active

        c_1 = self.c[matchups[:, 0], :]  # [bs, 2k]
        c_2 = self.c[matchups[:, 1], :]  # [bs, 2k]

        if use_cache:
            probs = self.cache['probs']
            dp_dc_2 = self.cache['c_1_times_omega']
        else:
            probs = self.predict(time_step=None, matchups=matchups, set_cache=False)
            dp_dc_2 = np.matmul(c_1, self.omega)

        per_match_diff = (outcomes - probs)[:, None]
        per_match_diff = np.hstack([per_match_diff, -per_match_diff])
        per_competitor_diff = (per_match_diff[:, :, None] * masks).sum(axis=(0, 1))

        dp_dc_1 = np.matmul(c_2, self.omega)
        dp_dc = np.stack([dp_dc_1, -1.0 * dp_dc_2], axis=1)
        c_updates = self.eta_c * per_match_diff[:, :, None] * dp_dc
        c_updates_pooled = (c_updates[:, :, None] * masks[:, :, :, None]).sum(axis=(0, 1))

        self.ratings[active_in_period] += self.eta_r * per_competitor_diff
        self.c[active_in_period] += c_updates_pooled

    def iterative_update(self, matchups, outcomes, use_cache=False, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            elo_diff = self.ratings[comp_1] - self.ratings[comp_2]
            c_1 = self.c[comp_1]
            c_2 = self.c[comp_2]

            melo_diff = np.dot(c_1, np.dot(self.omega, c_2)).item()
            prob = sigmoid(self.alpha * (elo_diff + melo_diff))
            delta = outcomes[idx] - prob

            rating_update = self.eta_r * delta

            self.ratings[comp_1] += rating_update
            self.ratings[comp_2] -= rating_update

            dp_dc_1 = np.dot(self.omega, c_2).flatten()
            dp_dc_2 = np.dot(self.omega, c_1).flatten()

            self.c[comp_1] += self.eta_c * delta * dp_dc_1
            self.c[comp_2] -= self.eta_c * delta * dp_dc_2



"""
Temporalized Massey Method
https://www.degruyter.com/document/doi/10.1515/jqas-2016-0093
https://arxiv.org/abs/1702.00585

"""
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.models.elo import Elo


class TemporalMassey(Elo):
    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1.0,
        alpha: float = 0.95,
        beta: float = 0.05,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        OnlineRatingSystem.__init__(self, competitors)
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        if (alpha is not None) and (beta is not None):
            self.alpha = alpha
            self.beta = beta
            self.get_coefs = self.get_constant_coefs
        else:
            self.get_coefs = self.get_varying_coefs
            self.matchup_counts = np.zeros(shape=self.num_competitors, dtype=dtype)

        self.cache = {'probs': None}
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.ratings[matchups]

    def get_constant_coefs(self, **kwargs):
        return self.alpha, self.beta, self.alpha, self.beta

    def get_varying_coefs_helper(self, idx):
        count = self.matchup_counts[idx]
        if count == 0:
            alpha, beta = 0.0, 0.0
        else:
            alpha = (count - 1.0) / count
            beta = 1.0 / count
        self.matchup_counts[idx] += 1
        return alpha, beta

    def get_varying_coefs(self, idx_1, idx_2):
        alpha_1, beta_1 = self.get_varying_coefs_helper(idx_1)
        alpha_2, beta_2 = self.get_varying_coefs_helper(idx_2)
        return alpha_1, beta_1, alpha_2, beta_2

    def predict(self, matchups: np.ndarray, set_cache: bool = False, **kwargs):
        """generate predictions"""
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        probs = ratings_1 / (ratings_1 + ratings_2)
        if set_cache:
            self.cache['probs'] = probs
        return probs

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """apply one update based on all of the results of the rating period"""
        pass

    def iterative_update(self, matchups, outcomes, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            r_1 = self.ratings[comp_1]
            r_2 = self.ratings[comp_2]
            a_1, b_1, a_2, b_2 = self.get_coefs(idx_1=comp_1, idx_2=comp_2)
            s_1 = (outcomes[idx] - 0.5) * 2.0
            s_2 = -1.0 * s_1
            r_1_new = (a_1 * r_1) + (b_1 * r_2) + (b_1 * s_1)
            r_2_new = (a_2 * r_2) + (b_2 * r_1) + (b_2 * s_2)
            self.ratings[comp_1] = r_1_new
            self.ratings[comp_2] = r_2_new



"""
example from: http://www.glicko.net/glicko/glicko2.pdf
"""
import pytest
import numpy as np
from riix.models.glicko2 import Glicko2
from riix.utils.data_utils import MatchupDataset


def test_glicko2():
    time_steps = np.array([0, 0, 0])  # all 3 matches are in the same time step
    matchups = np.array([[0, 1], [0, 2], [0, 3]])  # competitor 0 plays in 3 matchups
    outcomes = np.array([1.0, 0.0, 0.0])  # competitor 0 wins the first and loses the next 2
    dataset = MatchupDataset.init_from_arrays(
        time_steps=time_steps, matchups=matchups, outcomes=outcomes, competitors=[0, 1, 2, 3]
    )
    model = Glicko2(competitors=dataset.competitors, tau=0.5, update_method='batched')
    model.mus = (np.array([1500.0, 1400.0, 1550.0, 1700.0]) - 1500.0) / 173.7178
    model.phis = np.array([200.0, 30.0, 100.0, 300.0]) / 173.7178
    model.fit_dataset(dataset)
    # this is a really weak tolerance but alas Mr. Glickoman rounded to 4 decimal points at each step of the example
    assert model.mus[0] == pytest.approx(-0.2069, rel=5e-4)
    assert model.phis[0] == pytest.approx(0.8722, rel=5e-4)
    assert model.sigmas[0] == pytest.approx(0.05999, rel=5e-4)



"""classes and functions for working with rating data"""

import math
import time
from typing import List
import numpy as np
import polars as pl
from riix.utils.date_utils import get_duration

class MatchupDataset:
    """class for loading and iterating over paired comparison data"""

    def __init__(
        self,
        df: pl.DataFrame,
        competitor_cols: List[str],
        outcome_col: str,
        datetime_col: str = None,
        time_step_col: str = None,
        rating_period: str = '1W',
        verbose: bool = True,
    ):
        if len(competitor_cols) != 2:
            raise ValueError('must specify exactly 2 competitor columns')
        if (bool(datetime_col) + bool(time_step_col)) != 1:
            raise ValueError('must specify only one of time_step_col, datetime_col, timestamp_col')
        if time_step_col:
            self.time_steps = df[time_step_col]
        else:
            if df[datetime_col].dtype == pl.Datetime("ms"):
                datetime = df[datetime_col]
            elif df.schema[datetime_col] == pl.Date:
                datetime = df[datetime_col].cast(pl.Datetime)
            elif df.schema[datetime_col] == pl.Utf8:
                datetime = df.with_columns(
                    pl.when(pl.col(datetime_col).str.contains(r'^\d{4}-\d{2}-\d{2}$'))  # Match 'yyyy-mm-dd'
                    .then(pl.col(datetime_col).str.to_date('%Y-%m-%d'))
                    .otherwise(pl.col(datetime_col).str.strptime(pl.Datetime, '%Y-%m-%dT%H:%M:%S%.f', strict=False))
                )[datetime_col]
            else:
                raise ValueError('datetime_col must be one of Date, Datetime, or Utf8')
            seconds_since_epoch = (datetime.dt.timestamp() // 1_000_000).to_numpy()
            rating_period_duration_in_seconds = get_duration(rating_period)
            first_time = seconds_since_epoch[0]
            seconds_since_first_time = seconds_since_epoch - first_time
            self.time_steps = seconds_since_first_time // rating_period_duration_in_seconds

        self.time_steps = self.time_steps.astype(np.int32)
        self.process_time_steps()

        # Create a single competitors reference dataframe
        competitors_df = pl.DataFrame(
            {'competitor': pl.concat([
                df[competitor_cols[0]].cast(pl.Utf8).fill_null(''),
                df[competitor_cols[1]].cast(pl.Utf8).fill_null('')
            ]).unique().sort()
        }).lazy().select(
            pl.all(),
            pl.int_range(pl.len(), dtype=pl.Int32).alias('index')
        )
        self.competitors = sorted(competitors_df.collect()['competitor'].to_list())
        self.num_competitors = len(self.competitors)
        self.competitor_to_idx = dict(zip(self.competitors, range(self.num_competitors)))
        matchups_df = (df.lazy()
            .select([
                pl.col(competitor_cols[0]).cast(pl.Utf8).fill_null('').alias('comp1'),
                pl.col(competitor_cols[1]).cast(pl.Utf8).fill_null('').alias('comp2')
            ])
            .join(
                competitors_df,
                left_on='comp1',
                right_on='competitor'
            ).rename({'index': 'index1'})
            .join(
                competitors_df,
                left_on='comp2',
                right_on='competitor'
            ).rename({'index': 'index2'})
            .select(['index1', 'index2'])
        )
        self.matchups = np.ascontiguousarray(matchups_df.collect().to_numpy())



        self.outcomes = df[outcome_col].to_numpy()

        if verbose:
            print('loaded dataset with:')
            print(f'{self.matchups.shape[0]} matchups')
            print(f'{len(self.competitors)} unique competitors')
            print(f'{self.unique_time_steps.max() + 1} rating periods of length {rating_period}')

    def process_time_steps(self):
        self.unique_time_steps, unique_time_step_indices = np.unique(self.time_steps, return_index=True)
        self.time_step_end_idxs = np.roll(unique_time_step_indices, shift=-1)
        self.time_step_end_idxs[-1] = self.time_steps.shape[0]

    def __iter__(self):
        """iterate batches one rating period at a time"""
        period_start_idx = 0
        for time_step, period_end_idx in zip(self.unique_time_steps, self.time_step_end_idxs):
            matchups = self.matchups[period_start_idx:period_end_idx, :]
            outcomes = self.outcomes[period_start_idx:period_end_idx]
            period_start_idx = period_end_idx
            yield matchups, outcomes, time_step

    def __len__(self):
        return self.matchups.shape[0]

    def __getitem__(self, key):
        if isinstance(key, slice):
            slice_dataset = MatchupDataset.init_from_arrays(
                time_steps=self.time_steps[key],
                matchups=self.matchups[key, :],
                outcomes=self.outcomes[key],
                competitors=self.competitors,
            )
            slice_dataset.competitor_to_idx = self.competitor_to_idx
            return slice_dataset
        else:
            raise ValueError('you can only index MatchupDataset with a slice')

    @classmethod
    def init_from_arrays(cls, time_steps, matchups, outcomes, competitors):
        dataset = cls.__new__(cls)
        dataset.time_steps = time_steps
        dataset.process_time_steps()
        dataset.outcomes = outcomes
        dataset.matchups = matchups
        dataset.competitors = competitors
        dataset.num_competitors = len(competitors)
        return dataset

    @classmethod
    def load_from_npz(cls, path):
        time_steps, matchups, outcomes = np.load(path).values()
        competitors = np.unique(matchups)
        matchups = np.searchsorted(competitors, matchups)
        dataset = cls.init_from_arrays(
            time_steps=time_steps, matchups=matchups, outcomes=outcomes, competitors=competitors
        )
        print('loaded dataset with:')
        print(f'{dataset.matchups.shape[0]} matchups')
        print(f'{len(dataset.competitors)} unique competitors')
        return dataset


def split_matchup_dataset(dataset, test_fraction=0.2):
    split_idx = math.ceil(len(dataset) * (1.0 - test_fraction))
    train_dataset = dataset[:split_idx]
    test_dataset = dataset[split_idx:]
    print(f'split into train_dataset of length {len(train_dataset)} and test_dataset of length {len(test_dataset)}')
    return train_dataset, test_dataset

def generate_matchup_data(
    num_matchups: int = 10000,
    num_competitors: int = 100,
    num_rating_periods: int = 10,
    strength_var: float = 1.0,
    strength_noise_var: float = 0.1,
    theta: float = 1.001,
    seed: int = 0,
):
    start_time = int(time.time())
    matchups_per_period = num_matchups // num_rating_periods
    period_offsets = np.arange(num_rating_periods) * (3600 * 24)
    initial_timestamps = np.zeros(num_rating_periods, dtype=np.int64) + start_time
    timestamps = (initial_timestamps + period_offsets).repeat(matchups_per_period)

    rng = np.random.default_rng(seed=seed)
    strength_means = rng.normal(loc=0.0, scale=math.sqrt(strength_var), size=num_competitors)
    comp_1 = rng.integers(low=0, high=num_competitors, size=(num_matchups, 1))
    offset = rng.integers(low=1, high=num_competitors, size=(num_matchups, 1))
    comp_2 = np.mod(comp_1 + offset, num_competitors)
    matchups = np.hstack([comp_1, comp_2])
    strengths = strength_means[matchups]

    strength_noise = rng.normal(loc=0.0, scale=math.sqrt(strength_noise_var), size=strengths.shape)
    strengths = strengths + strength_noise
    # they need to be positive!
    strengths = np.exp(strengths)

    probs = np.zeros(shape=(num_matchups, 3))  # p(comp_1 win), p(draw), p(comp_2 win)
    probs[:, 0] = strengths[:, 0] / (strengths[:, 0] + (theta * strengths[:, 1]))
    probs[:, 2] = strengths[:, 1] / (strengths[:, 1] + (theta * strengths[:, 0]))
    probs[:, 1] = 1.0 - probs[:, 0] - probs[:, 2]

    outcomes = rng.multinomial(n=1, pvals=probs)
    outcomes = np.argmax(outcomes, axis=1) / 2.0  # map 0->0, 1->0.5, 2->1.0

    df = pl.DataFrame({
        'timestamp': timestamps,
        'competitor_1': matchups[:, 0],
        'competitor_2': matchups[:, 1],
        'outcome': outcomes,
    })
    df = (
        df.with_columns([
            (pl.col('timestamp') * 1000).cast(pl.Datetime('ms')).alias('date'),
            pl.concat_str([pl.lit('competitor_'), pl.col('competitor_1').cast(pl.Utf8)]).alias('competitor_1'),
            pl.concat_str([pl.lit('competitor_'), pl.col('competitor_2').cast(pl.Utf8)]).alias('competitor_2')
        ])
        .drop('timestamp')
    )
    return df



"""
The rating system by Cem Yuksel, 2024
  * paper: http://www.cemyuksel.com/research/matchmaking/i3d2024-matchmaking.pdf
  * supplemental: http://www.cemyuksel.com/research/matchmaking/i3d2024-matchmaking-supplemental.pdf
  * webpage: http://www.cemyuksel.com/research/matchmaking/
"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar
from riix.utils.constants import Q, Q2_3_OVER_PI2


class Yuksel2024(OnlineRatingSystem):
    """
    The rating system presented in Skill-Based Matchmaking for Competitive Two-Player Games by Cem Yuksel
    """

    rating_dim = 2

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1500.0,
        delta_r_max: float = 350.0,
        alpha: float = 2.0,
        scaling_factor: float = 0.9,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        """Initialize the rating system"""
        super().__init__(competitors)
        self.delta_r_max = delta_r_max
        self.alpha = alpha
        self.scaling_factor = scaling_factor
        self.ratings = np.zeros(shape=len(self.competitors), dtype=dtype) + initial_rating
        self.R = np.zeros(shape=len(self.competitors), dtype=dtype)
        self.W = np.zeros(shape=len(self.competitors), dtype=dtype) + 1e-2
        self.V = np.zeros(shape=len(self.competitors), dtype=dtype)
        self.D = np.zeros(shape=len(self.competitors), dtype=dtype)

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        """
        ratings = self.ratings[matchups]
        return sigmoid(Q * (ratings[:, 0] - ratings[:, 1]))

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        raise NotImplementedError

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
        """
        raise NotImplementedError

    @staticmethod
    def g_scalar(phi):
        return 1.0 / math.sqrt(1.0 + (Q2_3_OVER_PI2 * (phi**2.0)))

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
        """
        for (comp_1, comp_2), outcome in zip(matchups, outcomes):
            r_1 = self.ratings[comp_1]
            r_2 = self.ratings[comp_2]
            prob = sigmoid_scalar(Q * (r_1 - r_2))
            phi_1 = math.sqrt(self.V[comp_1] / self.W[comp_1])
            phi_2 = math.sqrt(self.V[comp_2] / self.W[comp_2])
            g_1 = self.g_scalar(phi_1)
            g_2 = self.g_scalar(phi_2)
            g_alpha_1 = self.g_scalar(self.alpha * phi_1)
            g_alpha_2 = self.g_scalar(self.alpha * phi_2)
            outcome_delta = outcome - prob
            F_1 = g_2 * outcome_delta
            F_2 = -g_1 * outcome_delta
            second_deriv = Q * prob * (1.0 - prob)
            D_1 = self.D[comp_1]
            D_1 = (g_2 * second_deriv) + (g_alpha_1 * D_1)
            D_2 = self.D[comp_2]
            D_2 = (g_1 * second_deriv) + (g_alpha_2 * D_2)
            delta_r = ((D_1 * F_1) - (D_2 * F_2)) / ((D_1**2.0) + (D_2**2.0))
            delta_r = min(self.delta_r_max, max(-self.delta_r_max, delta_r))
            if math.fabs(delta_r) != 0.0:
                D_1 = F_1 / delta_r
                D_2 = -F_2 / delta_r
            self.D[comp_1] = D_1
            self.D[comp_2] = D_2

            # UpdateStats 1
            scaled_update = self.scaling_factor * delta_r
            new_r_1 = r_1 + scaled_update
            self.ratings[comp_1] = new_r_1
            omega_1 = g_alpha_1
            self.W[comp_1] = omega_1 * self.W[comp_1] + 1.0
            delta_R_1 = new_r_1 - self.R[comp_1]
            self.R[comp_1] += delta_R_1 / self.W[comp_1]
            self.V[comp_1] = (omega_1 * self.V[comp_1]) + delta_R_1 * (new_r_1 - self.R[comp_1])

            # UpdateStats 2
            new_r_2 = r_2 - scaled_update
            self.ratings[comp_2] = new_r_2
            omega_2 = g_alpha_2
            self.W[comp_2] = omega_2 * self.W[comp_2] + 1.0
            delta_R_2 = new_r_2 - self.R[comp_2]
            self.R[comp_2] += delta_R_2 / self.W[comp_2]
            self.V[comp_2] = (omega_2 * self.V[comp_2]) + delta_R_2 * (new_r_2 - self.R[comp_2])

    def print_leaderboard(self, num_places):
        sorted_idxs = np.argsort(-self.ratings)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{self.ratings[comp_idx]:.6f}')



"""
A general class implementing online gradient based rating systems on differentiable likelihoods
"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
import jax.numpy as jnp
from jax import grad, jit
import jax.nn
from jax.scipy.stats import logistic


class AutogradRatingSystem(OnlineRatingSystem):
    """update the ratings using gradients"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        cdf: callable = logistic.cdf,
        scale: float = 400.0 / math.log(10.0),
        learning_rate: float = 32.0,
        initial_rating: float = 1500.0,
        update_method: str = 'iterative',
        dtype=jnp.float32,
    ):
        """Initialize the rating system"""
        super().__init__(competitors)

        self.ratings = jnp.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        self.cdf = cdf
        self.scale = scale

        @jit
        def likelihood_fn(ratings, matchups, outcomes):
            matchup_ratings = ratings[matchups]
            rating_diffs = matchup_ratings[:, 0] - matchup_ratings[:, 1]
            probs = self.cdf(rating_diffs / self.scale)
            return -jnp.log((outcomes * probs) + ((1.0 - outcomes) * (1.0 - probs))).sum()
        
        @jit
        def _predict_fn(ratings, matchups):
            matchup_ratings = ratings[matchups]
            rating_diffs = matchup_ratings[:, 0] - matchup_ratings[:, 1]
            probs = self.cdf(rating_diffs / self.scale)
            return probs
        self.predict_fn = _predict_fn
        

        self.grad_fn = grad(likelihood_fn)
        self.learning_rate = learning_rate * scale

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        """
        return self.predict_fn(self.ratings, matchups)

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.ratings[matchups]

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
        """
        grad = self.grad_fn(self.ratings, jnp.array(matchups), jnp.array(outcomes))
        update = grad * self.learning_rate
        self.ratings -= update

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
        """
        for idx in range(matchups.shape[0]):
            matchup = matchups[idx : idx + 1, :]
            outcome = outcomes[idx : idx + 1]
            grad = self.grad_fn(self.ratings, jnp.array(matchup), jnp.array(outcome))
            update = grad * self.learning_rate
            self.ratings -= update

    def print_leaderboard(self, num_places):
        sorted_idxs = np.argsort(-self.ratings)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{self.ratings[comp_idx]:.6f}')



"""
Models Module
=============

This module contains implementations of various rating systems used to evaluate and compare the performance of competitors in games or sports. Each system provides a unique approach to calculating ratings, taking into account factors such as game outcomes, player volatility, and more.

Included Rating Systems:
- Elo: A simple and widely used rating system that adjusts player ratings based on match outcomes.
- Elomentum: Enhances the Elo system by incorporating momentum into rating adjustments.
- Glicko: An advanced system extending Elo by introducing rating deviation and volatility.
- Iterative Markov: Uses Markov assumptions for iterative rating updates.
- Melo: A variation of the Elo rating system with multidimensional representations.
- Online Disc Decomposition: A machine learning approach for rating updates.
- SKF: Implements simplified Kalman filters for rating estimation.
- Temporal Massey: A temporal adaptation of the Massey rating system.
- TrueSkill: A Bayesian rating system developed by Microsoft.
- Velo: A variant of the Elo system with variance.
- Weng-Lin Bradley-Terry: A Bayesian online rating system using the Bradley-Terry (logistic) model.
- Weng-Lin Thurstone-Mosteller: A Bayesian online rating system using the Thurstone-Mosteller (gaussian) model.

Each rating system is implemented as a class with methods for initializing the system, updating player ratings based on game outcomes, and calculating expected scores between players. These classes are designed for use in online rating systems where ratings are updated continuously as games are played.

This module also provides utility functions and constants used across different rating systems, such as mathematical functions for calculating expected scores and scaling factors.

"""



"""
Bayesian generalization of Elo by Martin Ingram
https://www.researchgate.net/publication/348511584_How_to_extend_Elo_a_Bayesian_perspective
equations 17-19 are relevant for the case without adding surface, and best-of effects
"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar


class GenElo(OnlineRatingSystem):
    """Bayesian Generalized Elo from Ingram 2021"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_mu: float = 0.0,
        b: float = math.log(10.0) / 400.0,  # the temperature of the softmax
        sigma: float = 60.0,
        use_approx: bool = True,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        """Initialize the rating system"""
        super().__init__(competitors)
        self.mus = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_mu
        self.b = b
        self.b_over_2 = b / 2.0
        self.b_squared = b**2.0
        sigma_squared = sigma**2.0
        sigma_delta_squared = 2.0 * sigma_squared  # the var of the differece is the sum of the vars
        self.sigma_2_delta_inv = 1.0 / sigma_delta_squared
        if use_approx:
            self.alpha = math.sqrt(1.0 + ((math.pi * sigma_delta_squared * self.b_squared) / 8.0))
        else:
            self.alpha = 1.0

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        """
        mus = self.mus[matchups]
        mu_diffs = mus[:, 0] - mus[:, 1]
        return sigmoid(self.b * mu_diffs / self.alpha)

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.mus[matchups]

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
        """
        raise NotImplementedError

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
        """
        for idx, (comp_1, comp_2) in enumerate(matchups):
            mu_1 = self.mus[comp_1]
            mu_2 = self.mus[comp_2]
            prob = sigmoid_scalar(self.b * (mu_1 - mu_2))
            k = self.b_over_2 / ((self.sigma_2_delta_inv) + (self.b_squared * prob * (1.0 - prob)))
            delta = k * (outcomes[idx] - prob)
            self.mus[comp_1] += delta
            self.mus[comp_2] -= delta

    def print_leaderboard(self, num_places):
        raise NotImplementedError



"""simplified kalman filter"""
import math
import numpy as np
from scipy.stats import norm
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import base_10_sigmoid
from riix.utils.math_utils import v_and_w_win_scalar, v_and_w_draw_scalar, norm_cdf

LOG10 = math.log(10.0)
LOG10_SQUARED = LOG10**2.0


def bradley_terry_prob(z):
    return base_10_sigmoid(z)


def bradley_terry_scalar_prob_g_h(z, outcome):
    prob = base_10_sigmoid(z)
    g = LOG10 * (outcome - prob)
    h = LOG10_SQUARED * prob * (1.0 - prob)
    return prob, g, h


def thurstone_mosteller_scalar_prob_g_h(z, outcome):
    prob = norm_cdf(z)
    if outcome != 0.5:
        sign_multiplier = (2 * outcome) - 1  # maps 1 to 1 and 0 to -1
        v, w = v_and_w_win_scalar(z * sign_multiplier, 0.0)
        v = v * sign_multiplier
    else:
        v, w = v_and_w_draw_scalar(z, 0.0)
    return prob, v, w


class VSKF(OnlineRatingSystem):
    """vector covariance simplified kalman filter"""

    rating_dim = 2

    def __init__(
        self,
        competitors: list,
        mu_0: float = 0.0,
        v_0: float = 1.0,
        beta: float = 1.0,
        s: float = 1.0,
        epsilon: float = 1e-2,
        dtype=np.float64,
        model='bt',  # legal values are bt, tm, and dd
        update_method='iterative',
    ):
        super().__init__(competitors)
        self.mus = np.zeros(self.num_competitors, dtype=dtype) + mu_0
        self.vs = np.zeros(self.num_competitors, dtype=dtype) + v_0
        self.has_played = np.zeros(self.num_competitors, dtype=np.bool_)
        self.beta = beta
        self.beta2 = beta**2.0
        self.s = s
        self.s2 = s**2.0
        self.epsilon = epsilon
        self.prev_time_step = 0

        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

        if model == 'bt':
            self.predict_func = bradley_terry_prob
            self.prob_g_h_func = bradley_terry_scalar_prob_g_h
        elif model == 'tm':
            self.predict_func = norm.cdf
            self.prob_g_h_func = thurstone_mosteller_scalar_prob_g_h

    def predict(self, matchups: np.ndarray, set_cache: bool = False, **kwargs):
        """generate predictions"""
        ratings_1 = self.mus[matchups[:, 0]]
        ratings_2 = self.mus[matchups[:, 1]]
        rating_diffs = ratings_1 - ratings_2
        probs = self.predict_func(rating_diffs / self.s)
        return probs

    @property
    def ratings(self):
        return self.mus

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        mus = self.mus[matchups]
        vs = self.vs[matchups]
        ratings = np.concatenate((mus[..., None], vs[..., None]), axis=2).reshape(mus.shape[0], -1)
        return ratings

    def time_dynamics_update(self, time_step, matchups):
        """called once per period to model the increase in variance over time"""
        active_in_period = np.unique(matchups)
        self.has_played[active_in_period] = True
        time_delta = time_step - self.prev_time_step

        # update parameters for passage of time
        beta_t = self.beta**time_delta
        self.mus[self.has_played] = beta_t * self.mus[self.has_played]
        self.vs[self.has_played] = (beta_t**2) * self.vs[self.has_played] + (time_delta * self.epsilon)
        self.prev_time_step = time_step

        return active_in_period

    def batched_update(self, matchups, outcomes, time_step, use_cache=False):
        """apply one update based on all of the results of the rating period"""
        active_in_period = self.time_dynamics_update(time_step, matchups)
        masks = np.equal(matchups[:, :, None], active_in_period[None, :])  # N x 2 x active

        mus = self.mus[matchups]
        vs = self.vs[matchups]

        omegas = np.sum(vs, axis=1)
        probs = base_10_sigmoid((mus[:, 0] - mus[:, 1]) / self.s)

        g = LOG10 * (outcomes - probs)
        h = LOG10_SQUARED * probs * (1.0 - probs)

        denom = (self.s2) + h * omegas
        mu_updates = (self.s * g) / denom
        v_updates = h / denom

        mu_updates = mu_updates[:, None].repeat(2, 1)
        mu_updates[:, 1] *= -1

        v_updates = 1.0 - vs * v_updates[:, None]

        mu_updates_pooled = (mu_updates[:, :, None] * masks).sum(axis=(0, 1))
        v_updates_pooled = (v_updates[:, :, None] * masks).max(axis=(0, 1))

        self.mus[active_in_period] += self.vs[active_in_period] * mu_updates_pooled
        self.vs[active_in_period] *= v_updates_pooled

    def iterative_update(self, matchups, outcomes, time_step, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        self.time_dynamics_update(time_step, matchups)
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            mu_1 = self.mus[comp_1]
            mu_2 = self.mus[comp_2]
            v_1 = self.vs[comp_1]
            v_2 = self.vs[comp_2]
            outcome = outcomes[idx]

            omega = v_1 + v_2
            z = (mu_1 - mu_2) / self.s

            prob, g, h = self.prob_g_h_func(z, outcome)

            denom = (self.s2) + h * omega
            mu_update = (self.s * g) / denom

            v_update = h / denom

            self.mus[comp_1] += v_1 * mu_update
            self.mus[comp_2] -= v_2 * mu_update
            self.vs[comp_1] *= 1.0 - v_1 * v_update
            self.vs[comp_2] *= 1.0 - v_2 * v_update

    def print_leaderboard(self, num_places):
        sort_array = self.mus - (3.0 * np.sqrt(self.vs))
        sorted_idxs = np.argsort(-sort_array)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"mu - (3 * sd)"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{sort_array[comp_idx]:.6f}')



"""base class for online rating systems"""
from abc import ABC
from typing import Optional
import numpy as np
from riix.utils import MatchupDataset


class OnlineRatingSystem(ABC):
    """
    Base class for online rating systems. This class provides a framework for implementing
    various online rating systems, such as Elo, Glicko, or TrueSkill. It defines the basic
    structure and common methods that all such systems might share.

    Attributes:
        rating_dim (int): Dimension of competitor ratings. This could be 1 for systems like Elo,
                          where each competitor has a single rating value, or more for systems
                          like TrueSkill that use multiple values (e.g., mean and standard deviation).
        competitors (list): A list of competitors within the rating system.
        num_competitors (int): The number of competitors in the system.
    """

    rating_dim: int

    def __init__(self, competitors):
        """
        Initializes a new instance of an online rating system with a list of competitors.

        Parameters:
            competitors (list): A list of competitors to be included in the rating system. Each
                                competitor should have a structure or identifier compatible with
                                the specific rating system's requirements.
        """
        self.competitors = competitors
        self.num_competitors = len(competitors)

    def print_leaderboard(self, num_places=None):
        """
        Prints the leaderboard of the rating system.

        Parameters:
            num_places int: The number of top places to display on the leaderboard.
        """
        pass  # Implementation should be provided by subclasses.

    def predict(
        self,
        matchups: np.ndarray,
        time_step: int = None,
        set_cache: bool = False,
    ):
        raise NotImplementedError

    def update(self, matchups: np.ndarray, outcomes: np.ndarray, time_step: Optional[int], use_cache: bool = False):
        """
        Updates player ratings based on new matchup results.
        This general update function processes a set of matchups and their outcomes to adjust player ratings within the model.

        Parameters:
            matchups (np.ndarray): Array of matchups, where each matchup is represented by a pair of player indices
            outcomes (np.ndarray): Array of outcomes corresponding to each matchup represented as win (1), loss (0), or draw (0.5).
            time_step (int): The current time step or period of the rating update, used to adjust ratings over time.
            use_cache (bool, optional): Whether to use values cached during a prior call to predict() to speed up calculations. Defaults to False.
        """
        raise NotImplementedError

    def batched_update(self, matchups: np.ndarray, outcomes: np.ndarray, time_step: int, use_cache=False, **kwargs):
        """
        Performs a batched update of player ratings based on a series of matchups and their outcomes within a single time step.
        This method processes all matchups as occuring simultaneously, leveraging vectorized operations for efficiency

        Parameters:
            matchups (np.ndarray): Array of matchups, where each matchup is represented by a pair of player indices
            outcomes (np.ndarray): Array of outcomes corresponding to each matchup represented as win (1), loss (0), or draw (0.5).
            time_step (int): The current time step or period of the rating update, used to adjust ratings over time.
            use_cache (bool, optional): Whether to use values cached during a prior call to predict() to speed up calculations. Defaults to False.
        """
        raise NotImplementedError

    def iterative_update(self, matchups: np.ndarray, outcomes: np.ndarray, time_step: int, use_cache=False, **kwargs):
        """
        Updates player ratings iteratively for each matchup and outcome pair within a given time step treating them as if they were sequential
        This can be slower than batch processing but potentially more accurate by avoiding certain numerical issues

        Parameters:
            matchups (np.ndarray): Array of matchups, where each matchup is represented by a pair of player indices
            outcomes (np.ndarray): Array of outcomes corresponding to each matchup represented as win (1), loss (0), or draw (0.5).
            time_step (int): The current time step or period of the rating update, used to adjust ratings over time.
            use_cache (bool, optional): Whether to use values cached during a prior call to predict() to speed up calculations. Defaults to False.
        """
        raise NotImplementedError

    def get_pre_match_ratings(self, matchups: np.ndarray, time_step: Optional[int]) -> np.ndarray:
        """
        Returns the ratings for competitors at the timestep of the matchups
        Useful when using pre-match ratings as features in downstream ML pipelines

        Parameters:
            matchups (np.ndarray of shape (n,2)): competitor indices
            time_step (optional int)

        Returns:
            np.ndarray of shape (n,2): ratings for specified competitors
        """
        raise NotImplementedError

    def fit_batch(
        self,
        matchups: np.ndarray,
        outcomes: np.ndarray,
        time_step: int = None,
        return_pre_match_probs: bool = False,
        return_pre_match_ratings: bool = False,
        cache: bool = False,
    ) -> dict:
        if return_pre_match_probs:
            pre_match_probs = self.predict(matchups=matchups, time_step=time_step, set_cache=cache)
        if return_pre_match_ratings:
            pre_match_ratings = self.get_pre_match_ratings(matchups, time_step=time_step)
        self.update(matchups, outcomes, time_step=time_step, use_cache=cache)
        if return_pre_match_probs and return_pre_match_ratings:
            return pre_match_probs, pre_match_ratings
        elif return_pre_match_probs:
            return pre_match_probs
        elif return_pre_match_ratings:
            return_pre_match_ratings

    def fit_dataset(
        self,
        dataset: MatchupDataset,
        return_pre_match_probs: bool = False,
        return_pre_match_ratings: bool = False,
        cache: bool = False,
    ):
        """evaluate a rating system on a dataset"""
        n_matchups = len(dataset)
        if return_pre_match_probs:
            pre_match_probs = np.empty(shape=(n_matchups))
        if return_pre_match_ratings:
            pre_match_ratings = np.empty(shape=(n_matchups, 2 * self.rating_dim))

        idx = 0
        for matchups, outcomes, time_step in dataset:
            batch_outputs = self.fit_batch(
                matchups=matchups,
                outcomes=outcomes,
                time_step=time_step,
                return_pre_match_probs=return_pre_match_probs,
                return_pre_match_ratings=return_pre_match_ratings,
            )
            if (batch_outputs is not None) and (not isinstance(batch_outputs, tuple)):
                batch_outputs = (batch_outputs,)
            if return_pre_match_probs:
                batch_probs = batch_outputs[0]
                pre_match_probs[idx : idx + batch_probs.shape[0]] = batch_probs
            if return_pre_match_ratings:
                batch_pre_match_ratings = batch_outputs[-1]
                pre_match_ratings[idx : idx + batch_pre_match_ratings.shape[0]] = batch_pre_match_ratings
            if batch_outputs:
                idx += batch_outputs[0].shape[0]

        if return_pre_match_probs and return_pre_match_ratings:
            return pre_match_probs, pre_match_ratings
        elif return_pre_match_probs:
            return pre_match_probs
        elif return_pre_match_ratings:
            return_pre_match_ratings



import re

PROG = re.compile(r'(\d+)([WwDdHhMmSs])')
SECONDS_PER_UNIT = {
    'W': 7 * 24 * 60 * 60,    # weeks to seconds
    'D': 24 * 60 * 60,        # days to seconds
    'H': 60 * 60,             # hours to seconds
    'M': 60,                  # minutes to seconds
    'S': 1                    # seconds
}

def get_duration(duration_str):
    """
    Parse duration strings like '7D', '1W', '24H' etc. into the number os seconds since epoch as an int
    
    Parameters:
    -----------
    duration_str : str
        String in format numberLetter where Letter is one of:
        W/w - weeks
        D/d - days
        H/h - hours
        M/m - minutes
        S/s - seconds
        
    Returns:
    --------
    duration : int
    """
    # Extract number and unit using regex
    match = PROG.match(duration_str)
    if not match:
        raise ValueError(f"Invalid duration format: {duration_str}")
    number = int(match.group(1))
    unit = match.group(2).upper()
    duration = int(number * SECONDS_PER_UNIT[unit])
    return duration


"""utils for evaluating rating systems"""
import time
from functools import partial
from copy import deepcopy
from multiprocessing import Pool
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.utils.data_utils import MatchupDataset
from riix.metrics import binary_metrics_suite


def evaluate(model: OnlineRatingSystem, dataset: MatchupDataset, metrics_mask: np.ndarray=None):
    """evaluate a rating system on a dataset"""
    start_time = time.time()
    if metrics_mask is None:
        metrics_mask = np.ones(len(dataset), dtype=bool_)
    probs = model.fit_dataset(dataset, return_pre_match_probs=True)[metrics_mask]
    outcomes = dataset.outcomes[metrics_mask]
    duration = time.time() - start_time
    metrics = binary_metrics_suite(probs, outcomes)
    metrics['duration'] = duration
    return metrics


def eval_wrapper(params, rating_system_class, dataset, metrics_mask):
    model = rating_system_class(competitors=dataset.competitors, **params)
    return evaluate(model, dataset, metrics_mask)


def grid_search(
    rating_system_class,
    dataset,
    metrics_mask,
    param_configurations=None,
    metric='log_loss',
    minimize_metric=True,
    num_processes=None,
    seed=0,
    return_all_metrics=False,
):
    """Perform grid search and return the best hyperparameters."""
    map_fn = map
    if num_processes:
        pool = Pool(num_processes)
        map_fn = pool.map

    best_params = {}
    best_metrics = {}
    metric_multiplier = 1.0 if minimize_metric else -1.0
    best_metric = np.inf

    func = partial(
        eval_wrapper,
        rating_system_class=rating_system_class,
        dataset=dataset,
        metrics_mask=metrics_mask
    )
    all_metrics = list(map_fn(func, param_configurations))
    if num_processes:
        pool.close()
        pool.join()

    for current_params, current_metrics in zip(param_configurations, all_metrics):
        current_metric = current_metrics[metric]
        # Compare and update best metric and params
        if current_metric * metric_multiplier < best_metric:
            best_metric = current_metric * metric_multiplier
            best_metrics = deepcopy(current_metrics)
            best_params = current_params

    if not return_all_metrics:
        return best_params, best_metrics
    return best_params, best_metrics, all_metrics



"""
Elo + Momentum
Invented by Clayton Thorrez (me)
"""
import math
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.models.elo import Elo
from riix.utils.math_utils import sigmoid


class EloMentum(Elo):
    """Elo with momentum!"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1500.0,
        k: float = 32.0,
        alpha: float = math.log(10.0) / 400.0,
        momentum: float = 0.2,
        # momentum=(0.2, 0.9),
        momentum_type: str = 'nesterov',
        update_method: str = 'iterative',
        epsilon: float = 1e-8,
        dtype=np.float64,
    ):
        OnlineRatingSystem.__init__(self, competitors)
        self.k = k
        self.alpha = alpha
        self.momentum = momentum
        self.epsilon = epsilon
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating

        self.v = np.zeros(shape=self.num_competitors, dtype=dtype)
        self.cache = {'probs': None}
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

        if momentum_type == 'heavy_ball':
            self.momentum_fn = self.get_momentum_update
        elif momentum_type == 'nesterov':
            self.momentum_fn = self.get_nesterov_momentum_update
        elif momentum_type == 'adam':
            self.momentum_fn = self.get_adam_update
            self.mu = np.zeros(shape=self.num_competitors, dtype=dtype)
            self.t = np.zeros(shape=self.num_competitors, dtype=np.int32)
            self.beta1, self.beta2 = momentum
        else:
            raise ValueError(f'Invalid momentum_type {momentum_type}')

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.ratings[matchups]

    def predict(self, time_step: int, matchups: np.ndarray, set_cache: bool = False):
        """generate predictions"""
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        probs = sigmoid(self.alpha * (ratings_1 - ratings_2))
        if set_cache:
            self.cache['probs'] = probs
        return probs

    def update(self, matchups: np.ndarray, outcomes: np.ndarray, use_cache: bool = False, **kwargs):
        self.update(matchups, outcomes, use_cache=use_cache)

    def get_momentum_update(self, idx, g):
        v = self.v[idx]
        v_new = (self.momentum * v) + g
        self.v[idx] = v_new
        update = self.k * v_new
        return update

    def get_nesterov_momentum_update(self, idx, g):
        # from https://cs231n.github.io/neural-networks-3/#sgd
        # v_prev = v # back this up
        # v = mu * v - learning_rate * dx # velocity update stays the same
        # x += -mu * v_prev + (1 + mu) * v # position update changes form
        v = self.v[idx]
        v_new = (self.momentum * v) + (self.k * g)
        update = (self.momentum * v) + ((1.0 + self.momentum) * v_new)
        self.v[idx] = v_new
        return update

    def get_adam_update(self, idx, g):
        mu = self.mu[idx]
        v = self.v[idx]
        t = self.t[idx]
        t += 1.0
        mu_new = (self.beta1 * mu) + ((1.0 - self.beta1) * g)
        v_new = (self.beta2 * v) + ((1.0 - self.beta2) * (g * g))
        mu_hat = mu_new / (1.0 - (self.beta1**t))
        v_hat = v_new / (1.0 - (self.beta2**t))
        update = self.k * mu_hat / (math.sqrt(v_hat) + self.epsilon)
        self.mu[idx] = mu_new
        self.v[idx] = v_new
        self.t[idx] = t
        return update

    def batched_update(self, matchups, outcomes, use_cache):
        """apply one update based on all of the results of the rating period"""
        pass
        # active_in_period = np.unique(matchups)
        # masks = np.equal(matchups[:, :, None], active_in_period[None, :])  # N x 2 x active
        # if use_cache:
        #     probs = self.cache['probs']
        # else:
        #     probs = self.predict(time_step=None, matchups=matchups, set_cache=False)
        # per_match_diff = (outcomes - probs)[:, None]
        # per_match_diff = np.hstack([per_match_diff, -per_match_diff])
        # per_competitor_diff = (per_match_diff[:, :, None] * masks).sum(axis=(0, 1))
        # self.ratings[active_in_period] += self.k * per_competitor_diff

    def iterative_update(self, matchups, outcomes, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            diff = self.ratings[comp_1] - self.ratings[comp_2]
            prob = sigmoid(self.alpha * diff)
            # I think technically this is not the gradient because of alpha but it's close enough
            g_1 = outcomes[idx] - prob
            g_2 = -1.0 * g_1
            update_1 = self.momentum_fn(comp_1, g_1)
            update_2 = self.momentum_fn(comp_2, g_2)
            self.ratings[comp_1] += update_1
            self.ratings[comp_2] += update_2



"""mathematical constants computed once here to avoid recomputation"""
import math

# general math constants
PI2 = math.pi**2.0
THREE_OVER_PI_SQUARED = 3.0 / PI2

# glicko constants
Q = math.log(10.0) / 400.0
Q2 = Q**2.0
Q2_3 = 3.0 * Q2
Q2_3_OVER_PI2 = Q2_3 / PI2



"""Baseline rating systems"""
import numpy as np
from typing import Literal
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid


class BaselineRatingSystem(OnlineRatingSystem):
    """put a docstring"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        mode: Literal['win_rate', 'wins', 'appearances', 'random'],
        temperature: float = 1.0, # smoothing/scaling factor on the sigmoid
        prior: int = 2, # the number of virtual games to initialize competitors with, with 50% win rate to avoid divide by zero errors
        seed: int = 0
    ):
        """Initialize the rating system"""
        super().__init__(competitors)
        self.mode = mode
        self.temperature = temperature
        self.rng = np.random.default_rng(seed=seed)
        self.wins = np.zeros(shape=self.num_competitors) + (prior // 2)
        self.appearances = np.zeros(shape=self.num_competitors) + prior

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        Assumes a reparameterized Bradley-Terry model of p(i > j) = sigmoid(temperature * (strength_i - strength_j))
        The strengths are either based on count statistics such as wins, appearances, or win-rate, or randomly

        """
        if self.mode == 'win_rate':
            r_1 = self.wins[matchups[:,0]] / self.appearances[matchups[:,0]]
            r_2 = self.wins[matchups[:,1]] / self.appearances[matchups[:,1]]
        elif self.mode == 'wins':
            r_1 = self.wins[matchups[:,0]]
            r_2 = self.wins[matchups[:,1]]
        elif self.mode == 'appearances':
            r_1 = self.appearances[matchups[:,0]]
            r_2 = self.appearances[matchups[:,1]]
        else: # mode == 'random'
            r_1 = self.rng.uniform(size=matchups.shape[0])
            r_2 = self.rng.uniform(size=matchups.shape[0])
        return sigmoid(self.temperature * (r_1 - r_2))

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        raise NotImplementedError

    def update(self, matchups, outcomes, use_cache, **kwargs):
        """
        Apply a single update based on all results of the rating period, 

        Parameters:
            matchups: Matchup information for the rating period. Shape (N, 2)
            outcomes: Results of the matchups. Shape (N,)
        """
        active_in_period, active_indices = np.unique(matchups, return_inverse=True)
        active_matchups = active_indices.reshape(matchups.shape)
        
        # Update appearances
        appearance_counts = np.bincount(active_matchups.flatten(), minlength=len(active_in_period))
        self.appearances[active_in_period] += appearance_counts
        
        # Create masks for active competitors
        masks = (matchups[:, :, None] == active_in_period[None, None, :])  # Shape: (N, 2, len(active_in_period))
        
        # Calculate win counts
        win_contributions = np.stack([outcomes, 1 - outcomes], axis=1)[:, :, None]  # Shape: (N, 2, 1)
        active_win_counts = (win_contributions * masks).sum(axis=(0, 1))
        
        # Update wins
        self.wins[active_in_period] += active_win_counts
        
    def print_leaderboard(self, num_places):
        raise NotImplementedError



"""
Iterative Markov Method
https://www.degruyter.com/document/doi/10.1515/jqas-2019-0070
"""
import numpy as np
from riix.core.base import OnlineRatingSystem
from riix.models.elo import Elo


class IterativeMarkov(Elo):
    """Iterative Markov rating system"""

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 1.0,
        c: float = 0.1,
        weight_with_prob: bool = False,  # if True this becomes "Linear Elo"
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        OnlineRatingSystem.__init__(self, competitors)
        self.c = c
        self.weight_with_prob = weight_with_prob
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        self.cache = {'probs': None}
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.ratings[matchups]

    def predict(self, matchups: np.ndarray, set_cache: bool = False, **kwargs):
        """generate predictions"""
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        probs = ratings_1 / (ratings_1 + ratings_2)
        if set_cache:
            self.cache['probs'] = probs
        return probs

    def batched_update(self, matchups, outcomes, use_cache, **kwargs):
        """apply one update based on all of the results of the rating period"""
        pass

    def iterative_update(self, matchups, outcomes, use_cache=False, **kwargs):
        """treat the matchups in the rating period as if they were sequential"""
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            r_1 = self.ratings[comp_1]
            r_2 = self.ratings[comp_2]
            r_sum = r_1 + r_2
            prob_1 = r_1 / r_sum
            update = self.c * r_sum * (outcomes[idx] - prob_1)
            if self.weight_with_prob:
                update *= prob_1
            self.ratings[comp_1] += update
            self.ratings[comp_2] -= update



"""
example from: http://www.glicko.net/glicko/glicko.pdf
"""
import pytest
import numpy as np
from riix.models.glicko import Glicko
from riix.utils.data_utils import MatchupDataset


def test_glicko2():
    time_steps = np.array([0, 0, 0])  # all 3 matches are in the same time step
    matchups = np.array([[0, 1], [0, 2], [0, 3]])  # competitor 0 plays in 3 matchups
    outcomes = np.array([1.0, 0.0, 0.0])  # competitor 0 wins the first and loses the next 2
    dataset = MatchupDataset.init_from_arrays(
        time_steps=time_steps, matchups=matchups, outcomes=outcomes, competitors=[0, 1, 2, 3]
    )
    model = Glicko(competitors=dataset.competitors, c=0.0, update_method='batched')
    model.ratings = np.array([1500.0, 1400.0, 1550.0, 1700.0])
    model.rating_devs = np.array([200.0, 30.0, 100.0, 300.0])
    model.fit_dataset(dataset)
    # this is a really weak tolerance but alas Mr. Glickoman rounded to 4 decimal points at each step of the example
    assert model.ratings[0] == pytest.approx(1464.0, rel=1e-4)
    assert model.rating_devs[0] == pytest.approx(151.4, rel=1e-4)



"""The Online Rao Kupper rating system"""
import numpy as np
import math
from riix.core.base import OnlineRatingSystem
from riix.utils.math_utils import sigmoid, sigmoid_scalar


class OnlineRaoKupper(OnlineRatingSystem):
    """
    Implements an online version of the Rao Kupper model
    Performs stochastic gradient ascent on the log likelihood of the Rao Kupper model
    https://encyclopediaofmath.org/wiki/Paired_comparison_model
    https://papers.nips.cc/paper_files/paper/2015/hash/2a38a4a9316c49e5a833517c45d31070-Abstract.html
    """

    rating_dim = 1

    def __init__(
        self,
        competitors: list,
        initial_rating: float = 0.0,  # this does nothing
        theta: float = 2.0,
        step_size: float = 32.0,
        temperature: float = math.log(10.0) / 400.0,
        update_method: str = 'iterative',
        dtype=np.float64,
    ):
        """
        Initializes the Elo rating system with the given parameters.

        Parameters:
            competitors (list): A list of competitors to be rated within the system.
            initial_rating (float, optional): The initial Elo rating for new competitors. Defaults to 1500.0.
        """
        super().__init__(competitors)
        self.theta = theta
        self.log_theta = math.log(theta)
        self.step_size = step_size
        self.temperature = temperature
        self.ratings = np.zeros(shape=self.num_competitors, dtype=dtype) + initial_rating
        if update_method == 'batched':
            self.update = self.batched_update
        elif update_method == 'iterative':
            self.update = self.iterative_update

    def predict(self, matchups: np.ndarray, time_step: int = None, set_cache: bool = False):
        """
        Generates predictions for a series of matchups between competitors.
        Parameters:
            matchups (np.ndarray): A NumPy array of matchups, where each row represents a matchup
                                    and contains two integers indicating the indices of the competitors
                                    in the 'ratings' array.
            time_step (int, optional): A time step at which the predictions are made. This parameter
                                    is not used in the current implementation but can be utilized
                                    for time-dependent predictions. Defaults to None.
            set_cache (bool, optional): If True, caches the computed probabilities in the 'cache'
                                        attribute under the key 'probs'. Defaults to False.

        Returns:
            np.ndarray: A NumPy array containing the predicted probabilities for the first competitor
                        in each matchup winning against the second.
        """
        ratings_1 = self.ratings[matchups[:, 0]]
        ratings_2 = self.ratings[matchups[:, 1]]
        probs = sigmoid(self.temperature * (ratings_1 - ratings_2))
        return probs

    def get_pre_match_ratings(self, matchups: np.ndarray, **kwargs):
        return self.ratings[matchups]

    def batched_update(self, matchups, outcomes, use_cache):
        """
        Apply a single update based on all results of the rating period.

        Parameters:
            matchups: Matchup information for the rating period.
            outcomes: Results of the matchups.
            use_cache: Flag to use cached probabilities or calculate anew.
        """
        raise NotImplementedError

    def iterative_update(self, matchups, outcomes, **kwargs):
        """
        Treats the matchups in the rating period as sequential events.

        Parameters:
            matchups: Sequential matchups in the rating period.
            outcomes: Results of each matchup.
            **kwargs: Additional parameters (not used).
        """
        for idx in range(matchups.shape[0]):
            comp_1, comp_2 = matchups[idx]
            r_1 = self.ratings[comp_1]
            r_2 = self.ratings[comp_2]
            p_1 = sigmoid_scalar(self.temperature * (r_1 - r_2 - self.log_theta))
            p_2 = sigmoid_scalar(self.temperature * (r_2 - r_1 - self.log_theta))
            if outcomes[idx] != 0.5:
                g_1 = outcomes[idx] - p_1
                g_2 = (1.0 - outcomes[idx]) - p_2
            else:  # outcomes[idx] == 0.5
                # https://www.wolframalpha.com/input?i=derivative+of+log+%28%28exp%28a%29+*+exp%28b%29+*+%28theta%5E2+-+1%29%29+%2F+%28%28exp%28a%29+%2B+theta+*+exp%28b%29%29*%28exp%28b%29+%2B+theta*exp%28a%29%29%29%29+wrt+a
                g_1 = p_2 - p_1
                g_2 = p_1 - p_2

            self.ratings[comp_1] += self.step_size * g_1
            self.ratings[comp_2] += self.step_size * g_2

    def print_leaderboard(self, num_places):
        sorted_idxs = np.argsort(-self.ratings)[:num_places]
        max_len = min(np.max([len(comp) for comp in self.competitors] + [10]), 25)
        print(f'{"competitor": <{max_len}}\t{"rating"}')
        for p_idx in range(num_places):
            comp_idx = sorted_idxs[p_idx]
            print(f'{self.competitors[comp_idx]: <{max_len}}\t{self.ratings[comp_idx]:.6f}')



"""module for computing metrics for rating system experiments"""

import numpy as np


def binary_accuracy(probs: np.ndarray, outcomes: np.ndarray) -> float:
    """compute accuracy where outcomes is binary ties count for half"""
    pos_mask = probs > 0.5
    neg_mask = probs < 0.5
    draw_mask = probs == 0.5
    correct = outcomes[pos_mask].sum() + (1.0 - outcomes[neg_mask]).sum() + 0.5 * draw_mask.sum()
    return correct / probs.shape[0]

def accuracy_without_draws(probs: np.ndarray, outcomes: np.ndarray) -> float:
    """compute binary accuracy after first filtering out rows where the label is a draw"""
    draw_mask = outcomes == 0.5
    probs = probs[~draw_mask]
    outcomes = outcomes[~draw_mask]
    return binary_accuracy(probs, outcomes)


def accuracy_with_draws(probs: np.ndarray, outcomes: np.ndarray, draw_margin=0.0) -> float:
    """computes accuracy while allowing for ties"""
    pos_pred_mask = probs > (0.5 + draw_margin)
    neg_pred_mask = probs < (0.5 - draw_margin)
    draw_pred_mask = np.abs(probs - draw_margin) <= draw_margin
    correct = outcomes[pos_pred_mask].sum()
    correct += (1.0 - outcomes)[neg_pred_mask].sum()
    correct += 2 * outcomes[draw_pred_mask].sum()  # lmao
    return correct / outcomes.shape[0]


def binary_log_loss(probs: np.ndarray, outcomes: np.ndarray, eps: float = 1e-6) -> float:
    """compute log loss where outcome is binary 1.0 or 0.0"""
    probs = np.clip(probs, eps, 1 - eps)
    loss_array = -(np.log(probs) * outcomes) - (np.log(1.0 - probs) * (1.0 - outcomes))
    return loss_array.mean()


def brier_score(probs: np.ndarray, outcomes: np.ndarray) -> float:
    """compute the brier score, which is equivalent to the MSE"""
    return np.square(probs - outcomes).mean()


def binary_metrics_suite(probs: np.ndarray, outcomes: np.ndarray):
    """a wrapper class for running a bunch of binary metrics"""
    metrics = {
        'accuracy': binary_accuracy(probs, outcomes),
        'accuracy_without_draws' : accuracy_without_draws(probs, outcomes),
        'log_loss': binary_log_loss(probs, outcomes),
        'brier_score': brier_score(probs, outcomes),
    }
    return metrics
